searchData={"content_type":"text/markdown","items":[{"doc":"Spawn Elixir SDK is the support library for the Spawn Actors System.\nSpawn is a Stateful Serverless Platform for providing the multi-language Actor Model. For a broader understanding of Spawn, please consult its official [repository](https://github.com/eigr/spawn).\n\nThe advantage of the Elixir SDK over other SDKs is in Elixir's native ability to connect directly to an Erlang network. For this reason, the Elixir SDK allows any valid Elixir application to be part of a Spawn network without needing a sidecar attached.","ref":"SpawnSdk.html","title":"SpawnSdk","type":"module"},{"doc":"[Available in Hex](https://hex.pm/packages/spawn_sdk), the package can be installed\nby adding `spawn_sdk` and `spawn_statestores_*` to your list of dependencies in `mix.exs`:\n\n```elixir\ndef deps do\n  [\n    {:spawn_sdk, \"~> 1.0.0-rc.19\"},\n\n    # You can uncomment one of those dependencies if you are going to use Persistent Actors\n    #{:spawn_statestores_mariadb, \"~> 1.0.0-rc.19\"},\n    #{:spawn_statestores_mysql, \"~> 1.0.0-rc.19\"},\n    #{:spawn_statestores_postgres, \"~> 1.0.0-rc.19\"},\n    #{:spawn_statestores_mssql, \"~> 1.0.0-rc.19\"},\n    #{:spawn_statestores_cockroachdb, \"~> 1.0.0-rc.19\"},\n    #{:spawn_statestores_sqlite, \"~> 1.0.0-rc.19\"},\n  ]\nend\n```\n\n#","ref":"SpawnSdk.html#module-installation","title":"Installation - SpawnSdk","type":"module"},{"doc":"Following the steps below, you will have a valid Elixir application to use in a Spawn cluster. However, you will still need to generate a container image with your application to use it together with the [Spawn Operator for Kubernetes](https://github.com/eigr/spawn/tree/main/spawn_operator/spawn_operator).","ref":"SpawnSdk.html#module-deploy","title":"Deploy - SpawnSdk","type":"module"},{"doc":"After creating an Elixir application project, create the protobuf files for your business domain.\nIt is common practice to do this under the **priv/** folder. Let's demonstrate an example:\n\n```protobuf\nsyntax = \"proto3\";\n\npackage io.eigr.spawn.example;\n\nmessage MyState {\n  int32 value = 1;\n}\n\nmessage MyBusinessMessage {\n  int32 value = 1;\n}\n```\n\nIt is important to try to separate the type of message that must be stored as the actors' state from the type of messages\nthat will be exchanged between their actors' operations calls. In other words, the Actor's internal state is also represented\nas a protobuf type, and it is a good practice to separate these types of messages from the others in its business domain.\n\nIn the above case `MyState` is the type protobuf that represents the state of the Actor that we will create later\nwhile `MyBusiness` is the type of message that we will send and receive from this Actor.\n\nNow that we have defined our input and output types as Protobuf types we will need to compile these files to generate their respective Elixir modules. An example of how to do this can be found [here](https://github.com/eigr/spawn/blob/main/spawn_sdk/spawn_sdk_example/compile-example-pb.sh)\n\n> **_NOTE:_** You need to have installed the elixir plugin for protoc. More information on how to obtain and install the necessary tools can be found here [here](https://github.com/elixir-protobuf/protobuf#usage)\n\nNow that the protobuf types have been created we can proceed with the code. Example definition of an Actor.","ref":"SpawnSdk.html#module-how-to-use","title":"How to use - SpawnSdk","type":"module"},{"doc":"In this example we are creating an actor in a Named way, that is, it is a known actor at compile time.\n\n```elixir\ndefmodule SpawnSdkExample.Actors.MyActor do\n  use SpawnSdk.Actor,\n    name: \"jose\", # Default is Full Qualified Module name a.k.a __MODULE__\n    kind: :named, # Default is already :named. Valid are :named | :unamed | :pooled\n    stateful: true, # Default is already true\n    state_type: Io.Eigr.Spawn.Example.MyState, # or :json if you don't care about protobuf types\n    deactivate_timeout: 30_000,\n    snapshot_timeout: 2_000\n\n  require Logger\n\n  alias Io.Eigr.Spawn.Example.{MyState, MyBusinessMessage}\n\n  defact init(%Context{state: state} = ctx) do\n    Logger.info(\"[joe] Received InitRequest. Context: #{inspect(ctx)}\")\n\n    Value.of()\n    |> Value.state(state)\n  end\n\n  defact sum(\n           %MyBusinessMessage{value: value} = data,\n           %Context{state: state} = ctx\n         ) do\n    Logger.info(\"Received Request: #{inspect(data)}. Context: #{inspect(ctx)}\")\n\n    new_value = if is_nil(state), do: value, else: (state.value || 0) + value\n\n    Value.of(%MyBusinessMessage{value: new_value}, %MyState{value: new_value})\n  end\nend\n\n```\n\nWe declare two actions that the Actor can do. An initialization action that will be called every time an Actor instance is created and an action that will be responsible for performing a simple sum.\n\nNote Keep in mind that any Action that has the names present in the list below will behave as an initialization Action and will be called when the Actor is started (if there is more than one Action with one of these names, only one will be called).\n\nDefaults inicialization Action names: \"**init**\", \"**Init**\", \"**setup**\", \"**Setup**\"","ref":"SpawnSdk.html#module-named-actors","title":"Named Actors - SpawnSdk","type":"module"},{"doc":"We can also create Unnamed Dynamic/Lazy actors, that is, despite having its unamed behavior defined at compile time, a Lazy actor will only have a concrete instance when it is associated with an identifier/name at runtime. Below follows the same previous actor being defined as Unamed.\n\n```elixir\ndefmodule SpawnSdkExample.Actors.UnamedActor do\n  use SpawnSdk.Actor,\n    name: \"unamed_actor\",\n    kind: :unamed,\n    state_type: Io.Eigr.Spawn.Example.MyState\n\n  require Logger\n\n  alias Io.Eigr.Spawn.Example.{MyState, MyBusinessMessage}\n\n  defact sum(\n           %MyBusinessMessage{value: value} = data,\n           %Context{state: state} = ctx\n         ) do\n    Logger.info(\"Received Request: #{inspect(data)}. Context: #{inspect(ctx)}\")\n\n    new_value = if is_nil(state), do: value, else: (state.value || 0) + value\n\n    Value.of(%MyBusinessMessage{value: new_value}, %MyState{value: new_value})\n  end\nend\n```\n\nNotice that the only thing that has changed is the the kind of actor, in this case the kind is set to :unamed.\n\n> **_NOTE:_** Can Elixir programmers think in terms of Named vs Unamed actors as more or less known at startup vs dynamically supervised/registered? That is, defining your actors directly in the supervision tree or using a Dynamic Supervisor for that.","ref":"SpawnSdk.html#module-unamed-actor","title":"Unamed Actor - SpawnSdk","type":"module"},{"doc":"Sometimes we want a particular actor to be able to serve requests concurrently,\nhowever actors will always serve one request at a time using buffering mechanisms to receive requests in their mailbox and serve each request one by one.\nSo to get around this behaviour you can configure your Actor as a Pooled Actor, this way the system will generate a pool of actors to meet certain requests. See an example below:\n\n```elixir\ndefmodule SpawnSdkExample.Actors.PooledActor do\n  use SpawnSdk.Actor,\n    name: \"pooled_actor\",\n    kind: :pooled,\n    stateful: false\n\n  require Logger\n\n  defact ping(_data, %Context{} = ctx) do\n    Logger.info(\"Received Request. Context: #{inspect(ctx)}\")\n\n    Value.void()\n  end\nend\n```\n\n> **_NOTE:_** It is important to mention that Pooled Actors cannot be stateful because concurrent accesses for Stateful actors could generate undesired inconsistencies.\n\nTo invoke a pooled actor, it is necessary to pass the pooled: true option to the invoke function. See an example:\n\n```elixir\niex(spawn_a1@127.0.0.1)16> for _n   for _n   ]:[notice]:Activating Actor pooled_actor-5 with Parent pooled_actor-5 in Node :\"spawn_a1@127.0.0.1\". Persistence false.\n2022-11-25 11:21:53.759 [spawn_a1@127.0.0.1]:[pid=  ]:[debug]:Actor pooled_actor-5 reactivated. ActorRef PID: #PID \n2022-11-25 11:21:53.760 [spawn_a1@127.0.0.1]:[pid=  ]:[info]:Received Request. Context: %SpawnSdk.Context{state: nil, caller: nil, self: %Eigr.Functions.Protocol.Actors.ActorId{name: \"pooled_actor-5\", system: \"spawn-system\", parent: \"pooled_actor-5\", __unknown_fields__: []}, metadata: %{}}\n2022-11-25 11:21:53.763 [spawn_a1@127.0.0.1]:[pid=  ]:[notice]:Activating Actor pooled_actor-4 with Parent pooled_actor-4 in Node :\"spawn_a1@127.0.0.1\". Persistence false.\n2022-11-25 11:21:53.763 [spawn_a1@127.0.0.1]:[pid=  ]:[debug]:Actor pooled_actor-4 reactivated. ActorRef PID: #PID \n2022-11-25 11:21:53.765 [spawn_a1@127.0.0.1]:[pid=  ]:[info]:Received Request. Context: %SpawnSdk.Context{state: nil, caller: nil, self: %Eigr.Functions.Protocol.Actors.ActorId{name: \"pooled_actor-4\", system: \"spawn-system\", parent: \"pooled_actor-4\", __unknown_fields__: []}, metadata: %{}}\n2022-11-25 11:21:53.767 [spawn_a1@127.0.0.1]:[pid=  ]:[notice]:Activating Actor pooled_actor-7 with Parent pooled_actor-7 in Node :\"spawn_a1@127.0.0.1\". Persistence false.\n2022-11-25 11:21:53.767 [spawn_a1@127.0.0.1]:[pid=  ]:[debug]:Actor pooled_actor-7 reactivated. ActorRef PID: #PID \n2022-11-25 11:21:53.768 [spawn_a1@127.0.0.1]:[pid=  ]:[info]:Received Request. Context: %SpawnSdk.Context{state: nil, caller: nil, self: %Eigr.Functions.Protocol.Actors.ActorId{name: \"pooled_actor-7\", system: \"spawn-system\", parent: \"pooled_actor-7\", __unknown_fields__: []}, metadata: %{}}\n2022-11-25 11:21:53.770 [spawn_a1@127.0.0.1]:[pid=  ]:[notice]:Activating Actor pooled_actor-2 with Parent pooled_actor-2 in Node :\"spawn_a1@127.0.0.1\". Persistence false.\n2022-11-25 11:21:53.771 [spawn_a1@127.0.0.1]:[pid=  ]:[debug]:Actor pooled_actor-2 reactivated. ActorRef PID: #PID \n2022-11-25 11:21:53.772 [spawn_a1@127.0.0.1]:[pid=  ]:[info]:Received Request. Context: %SpawnSdk.Context{state: nil, caller: nil, self: %Eigr.Functions.Protocol.Actors.ActorId{name: \"pooled_actor-2\", system: \"spawn-system\", parent: \"pooled_actor-2\", __unknown_fields__: []}, metadata: %{}}\n2022-11-25 11:21:53.774 [spawn_a1@127.0.0.1]:[pid=  ]:[notice]:Activating Actor pooled_actor-8 with Parent pooled_actor-8 in Node :\"spawn_a1@127.0.0.1\". Persistence false.\n2022-11-25 11:21:53.775 [spawn_a1@127.0.0.1]:[pid=  ]:[debug]:Actor pooled_actor-8 reactivated. ActorRef PID: #PID \n2022-11-25 11:21:53.776 [spawn_a1@127.0.0.1]:[pid=  ]:[info]:Received Request. Context: %SpawnSdk.Context{state: nil, caller: nil, self: %Eigr.Functions.Protocol.Actors.ActorId{name: \"pooled_actor-8\", system: \"spawn-system\", parent: \"pooled_actor-8\", __unknown_fields__: []}, metadata: %{}}\n2022-11-25 11:21:53.777 [spawn_a1@127.0.0.1]:[pid=  ]:[notice]:Activating Actor pooled_actor-9 with Parent pooled_actor-9 in Node :\"spawn_a1@127.0.0.1\". Persistence false.\n2022-11-25 11:21:53.778 [spawn_a1@127.0.0.1]:[pid=  ]:[debug]:Actor pooled_actor-9 reactivated. ActorRef PID: #PID \n2022-11-25 11:21:53.778 [spawn_a1@127.0.0.1]:[pid=  ]:[info]:Received Request. Context: %SpawnSdk.Context{state: nil, caller: nil, self: %Eigr.Functions.Protocol.Actors.ActorId{name: \"pooled_actor-9\", system: \"spawn-system\", parent: \"pooled_actor-9\", __unknown_fields__: []}, metadata: %{}}\n2022-11-25 11:21:53.779 [spawn_a1@127.0.0.1]:[pid=  ]:[debug]:Lookup Actor pooled_actor. PID: #PID \n2022-11-25 11:21:53.780 [spawn_a1@127.0.0.1]:[pid=  ]:[info]:Received Request. Context: %SpawnSdk.Context{state: nil, caller: nil, self: %Eigr.Functions.Protocol.Actors.ActorId{name: \"pooled_actor-8\", system: \"spawn-system\", parent: \"pooled_actor-8\", __unknown_fields__: []}, metadata: %{}}\n2022-11-25 11:21:53.780 [spawn_a1@127.0.0.1]:[pid=  ]:[debug]:Lookup Actor pooled_actor. PID: #PID \n2022-11-25 11:21:53.781 [spawn_a1@127.0.0.1]:[pid=  ]:[info]:Received Request. Context: %SpawnSdk.Context{state: nil, caller: nil, self: %Eigr.Functions.Protocol.Actors.ActorId{name: \"pooled_actor-5\", system: \"spawn-system\", parent: \"pooled_actor-5\", __unknown_fields__: []}, metadata: %{}}\n2022-11-25 11:21:53.781 [spawn_a1@127.0.0.1]:[pid=  ]:[debug]:Lookup Actor pooled_actor. PID: #PID \n[\n  ok: nil,\n  ok: nil,\n  ok: nil,\n  ok: nil,\n  ok: nil,\n  ok: nil,\n  ok: nil,\n  ok: nil,\n  ok: nil\n]\niex(spawn_a1@127.0.0.1)17> 2022-11-25 11:21:53.782 [spawn_a1@127.0.0.1]:[pid=  ]:[info]:Received Request. Context: %SpawnSdk.Context{state: nil, caller: nil, self: %Eigr.Functions.Protocol.Actors.ActorId{name: \"pooled_actor-4\", system: \"spawn-system\", parent: \"pooled_actor-4\", __unknown_fields__: []}, metadata: %{}}\n2022-11-25 11:22:03.999 [spawn_a1@127.0.0.1]:[pid=  ]:[debug]:Deactivating actor pooled_actor-9 for timeout\n2022-11-25 11:22:04.226 [spawn_a1@127.0.0.1]:[pid=  ]:[debug]:Deactivating actor pooled_actor-5 for timeout\n2022-11-25 11:22:07.271 [spawn_a1@127.0.0.1]:[pid=  ]:[debug]:Deactivating actor pooled_actor-8 for timeout\n2022-11-25 11:22:10.167 [spawn_a1@127.0.0.1]:[pid=  ]:[debug]:Deactivating actor pooled_actor-7 for timeout\n2022-11-25 11:22:12.038 [spawn_a1@127.0.0.1]:[pid=  ]:[debug]:Deactivating actor pooled_actor-4 for timeout\n2022-11-25 11:22:12.613 [spawn_a1@127.0.0.1]:[pid=  ]:[debug]:Deactivating actor pooled_actor-2 for timeout\n```\n\nIt is possible to configure the minimum and maximum pool size by passing the min_pool_size and max_pool_size options. The actor below would create a pool of a maximum of 100 Actors:\n\n```elixir\ndefmodule SpawnSdkExample.Actors.PooledActor do\n  use SpawnSdk.Actor,\n    name: \"pooled_actor\",\n    kind: :pooled,\n    min_pool_size: 1,\n    max_pool_size: 10,\n    stateful: false\n\n  # Code omitted for brevity\nend\n```\n\n> **_ALERT:_** Keep in mind that a large pool size does not always mean better performance. In fact, smaller pool sizes sometimes often have a much greater effect in this regard. It is interesting that you experiment with different pool sizes until you find the one that best suits your workload.","ref":"SpawnSdk.html#module-pooled-actors","title":"Pooled Actors - SpawnSdk","type":"module"},{"doc":"Actors can also emit side effects to other Actors as part of their response. See an example:\n\n```elixir\ndefmodule SpawnSdkExample.Actors.UnamedActor do\n  use SpawnSdk.Actor,\n    kind: :unamed,\n    stateful: false,\n    state_type: Io.Eigr.Spawn.Example.MyState\n\n  require Logger\n\n  alias Io.Eigr.Spawn.Example.{MyState, MyBusinessMessage}\n\n  alias SpawnSdk.Flow.SideEffect\n\n  defact sum(%MyBusinessMessage{value: value} = data, %Context{state: state} = ctx) do\n    Logger.info(\"Received Request: #{inspect(data)}. Context: #{inspect(ctx)}\")\n\n    new_value = if is_nil(state), do: value, else: (state.value || 0) + value\n\n    result = %MyBusinessMessage{value: new_value}\n    new_state = %MyState{value: new_value}\n\n    Value.of()\n    |> Value.response(result)\n    |> Value.state(new_state)\n    |> Value.effects(\n      # This returns a list of side effects. In this case containing only one effect. However, multiple effects can be chained together,\n      # just by calling the effect function as shown here.\n      # The delay means that it will be fired asynchronously after 5000 milliseconds (5 seconds)\n      # If only one effect is desired, you can also choose to use the to/3 function together with Value.effect().\n      # Example: Values.effect(SideEffect.to(name, func, payload))\n      SideEffect.of()\n      |> SideEffect.effect(\"joe\", :sum, result, delay: 5_000, scheduled_to: ~U[2020-01-01 10:00:00.145Z])\n      # use delay or scheduled_to, not both\n    )\n  end\nend\n\n```\n\nIn the example above we see that the Actor joe will receive a request as a side effect from the Actor who issued this effect.\n\nSide effects do not interfere with an actor's request-response flow. They will \"always\" be processed asynchronously and any response sent back from the Actor receiving the effect will be ignored by the effector.","ref":"SpawnSdk.html#module-side-effects","title":"Side Effects - SpawnSdk","type":"module"},{"doc":"Actors can also route some actions to other actors as part of their response. See an example:\n\n```elixir\ndefmodule SpawnSdkExample.Actors.ForwardPipeActor do\n  use SpawnSdk.Actor,\n    name: \"pipeforward\",\n    kind: :named,\n    stateful: false\n\n  require Logger\n\n  alias Io.Eigr.Spawn.Example.MyBusinessMessage\n\n  defact forward_example(%MyBusinessMessage{} = msg, _ctx) do\n    Logger.info(\"Received request with #{msg.value}\")\n\n    Value.of()\n    |> Value.forward(\n      Forward.to(\"second_actor\", \"sum_plus_one\")\n    )\n    |> Value.void()\n  end\n\n  defact pipe_example(%MyBusinessMessage{} = msg, _ctx) do\n    Logger.info(\"Received request with #{msg.value}\")\n\n    Value.of()\n    |> Value.response(MyBusinessMessage.new(value: 999))\n    |> Value.pipe(\n      Pipe.to(\"second_actor\", \"sum_plus_one\")\n    )\n    |> Value.void()\n  end\nend\n\ndefmodule SpawnSdkExample.Actors.SecondActorExample do\n  use SpawnSdk.Actor,\n    name: \"second_actor\",\n    stateful: false\n\n  require Logger\n\n  alias Io.Eigr.Spawn.Example.MyBusinessMessage\n\n  defact sum_plus_one(%MyBusinessMessage{} = msg, _ctx) do\n    Logger.info(\"Received request with #{msg.value}\")\n\n    Value.of()\n    |> Value.response(MyBusinessMessage.new(value: msg.value + 1))\n    |> Value.void()\n  end\nend\n\n```\n\nWe are returning void in both examples so we dont care about what is being stored in the actor state.\n\nIn the case above, every time you call the `forward_example` the second_actor's `sum_plus_one` function will receive the value forwarded originally in the invocation as its input. The end result will be:\n\n```elixir\niex> SpawnSdk.invoke(\"pipeforward\", system: \"spawn-system\", action: \"forward_example\", payload: %Io.Eigr.Spawn.Example.MyBusinessMessage{value: 1})\n{:ok, %Io.Eigr.Spawn.Example.MyBusinessMessage{value: 2}}\n```\n\nFor the Pipe example, the the second_actor's `sum_plus_one` function will always receive `%MyBusinessMessage{value: 999}` due to getting the value from the previous specification in the `pipe_example` action, the end result will be:\n\n```elixir\niex> SpawnSdk.invoke(\"pipeforward\", system: \"spawn-system\", action: \"pipe_example\", payload: %Io.Eigr.Spawn.Example.MyBusinessMessage{value: 1})\n{:ok, %Io.Eigr.Spawn.Example.MyBusinessMessage{value: 1000}}\n```","ref":"SpawnSdk.html#module-pipe-and-forward","title":"Pipe and Forward - SpawnSdk","type":"module"},{"doc":"Actors can also send messages to a group of actors at once as an action callback. See the example below:\n\n```elixir\ndefmodule Fleet.Actors.Driver do\n  use SpawnSdk.Actor,\n    kind: :unamed,\n    # Set ´driver´ channel for all actors of the same type (Fleet.Actors.Driver)\n    channel: \"drivers\",\n    state_type: Fleet.Domain.Driver\n\n  alias Fleet.Domain.{\n    Driver,\n    OfferRequest,\n    OfferResponse,\n    Point\n  }\n\n  require Logger\n\n  @brain_actor_channel \"fleet-controllers\"\n\n  defact update_position(%Point{} = position, %Context{state: %Driver{id: name} = driver} = ctx) do\n    Logger.info(\n      \"Driver [#{name}] Received Update Position Event. Position: [#{inspect(position)}]. Context: #{inspect(ctx)}\"\n    )\n\n    driver_state = %Driver{driver | position: position}\n\n    %Value{}\n    |> Value.of(driver_state, driver_state)\n    |> Value.broadcast(\n      Broadcast.to(\n        @brain_actor_channel,\n        \"driver_position\",\n        driver_state\n      )\n    )\n  end\nend\n```\n\nIn the case above, every time an Actor \"driver\" executes the update_position action it will send a message to all the actors participating in the channel called \"fleet-controllers\".\n\n#","ref":"SpawnSdk.html#module-broadcast","title":"Broadcast - SpawnSdk","type":"module"},{"doc":"Sometimes you may want to send events out of ActorSystem using Phoenix.PubSub.\nOne way to do this is to take advantage of the same Broadcast infrastructure that Spawn offers you but indicating an external channel. Below is an example:\n\n1. Create a Listener to receive the events using the `SpawnSdk.Channel.Subscriber` helper module.\n\n```elixir\ndefmodule SpawnSdkExample.Subscriber do\n  @moduledoc \"\"\"\n  This module exemplifies how to listen for pubsub events that were emitted by actors but that will be treated not by actors but as normal pubsub events.\n  This is particularly useful for integrations between Spawn and Phoenix LiveView.\n  \"\"\"\n  use GenServer\n  require logger\n\n  alias SpawnSdk.Channel.Subscriber\n\n  @impl true\n  define init(state) do\n    Subscriber.subscribe(\"external.channel\")\n    {:ok, state}\n  end\n\n  @impl true\n  def handle_info({:receive, payload}, state) do\n    Logger.info(\"Received pubsub event #{inspect(payload)}\")\n    {:noreply, state}\n  end\n\n  def start_link(args) do\n    GenServer.start_link(__MODULE__, args)\n  end\nend\n```\n\nYou need to match using the {:receive, payload} tuple in your handle_info.\n\n> **_NOTE:_** By default SpawnSdk.Channel.Subscriber will subscribe to pubsub using the atom :actor_channel as an argument.\n\n    If you need to change this, just configure your configuration as follows:\n\n**_config.exs_**\n\n```elixir\nconfig :spawn,\n   pubsub_group: :your_channel_group_here\n```\n\n2. SpawnSdk.System.Supervisor.\n\n```elixir\ndefmodule SpawnSdkExample.Application do\n  @moduledoc false\n  use Application\n\n  @impl true\n  def start(_type, _args) do\n    children = [\n      {\n        SpawnSdk.System.Supervisor,\n        system: \"spawn-system\",\n        actors: [\n          SpawnSdkExample.Actors.JoeActor\n        ],\n        extenal_subscribers: [\n          {SpawnSdkExample.Subscriber, []}\n        ]\n      }\n    ]\n\n    opts = [strategy: :one_for_one, name: SpawnSdkExample.Supervisor]\n    Supervisor.start_link(children, opts)\n  end\nend\n```\n\nThe important thing here is to use the external_subscribers attribute. As seen above :external_subscribers accepts a list of specs as a parameter.\n\n3. Set your actor as you normally would and emit your broadcast events using Broadcast.to(channel, payload).\n\n```elixir\ndefmodule SpawnSdkExample.Actors.JoeActor do\n  use SpawnSdk.Actor,\n    name: \"joe\",\n    state_type: Io.Eigr.Spawn.Example.MyState\n\n  require Logger\n  alias Io.Eigr.Spawn.Example.{MyState, MyBusinessMessage}\n\n  defact sum(%MyBusinessMessage{value: value} = data, %Context{state: state} = ctx) do\n    Logger.info(\"[joe] Received Request: #{inspect(data)}. Context: #{inspect(ctx)}\")\n\n    new_value =\n      if is_nil(state) do\n        0 + value\n      else\n        (state.value || 0) + value\n      end\n\n    response = %MyBusinessMessage{value: new_value}\n\n    %Value{}\n    |> Value.of(response, %MyState{value: new_value})\n    |> Value.broadcast(Broadcast.to(\"my.channel\", response))\n  end\nend\n```","ref":"SpawnSdk.html#module-broadcast-to-external-subscribers","title":"Broadcast to External Subscribers - SpawnSdk","type":"module"},{"doc":"Actors can also declare Actions that act recursively as timers. See an example below:\n\n```elixir\ndefmodule SpawnSdkExample.Actors.ClockActor do\n  use SpawnSdk.Actor,\n    name: \"clock_actor\",\n    state_type: Io.Eigr.Spawn.Example.MyState,\n    deactivate_timeout: 86_400_000\n\n  require Logger\n\n  alias Io.Eigr.Spawn.Example.MyState\n\n  @set_timer 15_000\n  defact clock(%Context{state: state} = ctx) do\n    Logger.info(\"[clock] Clock Actor Received Request. Context: #{inspect(ctx)}\")\n\n    new_value = if is_nil(state), do: 0, else: state.value + 1\n    new_state = MyState.new(value: new_value)\n\n    Value.of()\n    |> Value.state(new_state)\n  end\nend\n```\n\n> **_NOTE:_** Timers Actions are recorded as Actor metadata. Where in turn we use a synchronization mechanism via CRDTs to keep the metadata alive in the cluster while there is an active Spawn VM. That is, Timers Actions are ephemeral and therefore only exist while there is at least one active VM in the cluster.\n\nIn the example above the ´clock´ action will be called every 15 seconds.","ref":"SpawnSdk.html#module-timers","title":"Timers - SpawnSdk","type":"module"},{"doc":"Once we define our actors we can now declare our supervision tree:\n\n```elixir\ndefmodule SpawnSdkExample.Application do\n  @moduledoc false\n  use Application\n\n  @impl true\n  def start(_type, _args) do\n    children = [\n      {\n        SpawnSdk.System.Supervisor,\n        system: \"spawn-system\",\n        actors: [\n          SpawnSdkExample.Actors.MyActor,\n          SpawnSdkExample.Actors.UnamedActor,\n          SpawnSdkExample.Actors.ClockActor,\n          SpawnSdkExample.Actors.PooledActor\n        ]\n      }\n    ]\n\n    opts = [strategy: :one_for_one, name: SpawnSdkExample.Supervisor]\n    Supervisor.start_link(children, opts)\n  end\nend\n```","ref":"SpawnSdk.html#module-declaring-the-supervision-tree","title":"Declaring the supervision tree - SpawnSdk","type":"module"},{"doc":"Actors also have some standard actions that are not implemented by the user and that can be used as a way to get the state of an actor without the invocation requiring an extra trip to the host functions. You can think of them as a cache of their state, every time you invoke a default action on an actor it will return the value directly from the Sidecar process without this target process needing to invoke its equivalent host function.\n\nLet's take an example. Suppose Actor Joe wants to know the current state of Actor Robert. What Joe can do is invoke Actor Robert's default action called get_state. This will make Actor Joe's sidecar find Actor Robert's sidecar somewhere in the cluster and Actor Robert's sidecar will return its own state directly to Joe without having to resort to your host function, this in turn will save you a called over the network and therefore this type of invocation is faster than invocations of user-defined actions usually are.\n\nAny invocations to actions with the following names will follow this rule: \"**get**\",\n\"**Get**\",\n\"**get_state**\",\n\"**getState**\",\n\"**GetState**\"\n\n> **_NOTE:_** You can override this behavior by defining your actor as an action with the same name as the default actions. In this case it will be the Action defined by you that will be called, implying perhaps another network roundtrip","ref":"SpawnSdk.html#module-default-actions","title":"Default Actions - SpawnSdk","type":"module"},{"doc":"To deploy your actors in a cluster you need to use our Controller for Kubernetes. Detailed information on how to proceed can be found [here](https://github.com/eigr/spawn#install) and [here](https://github.com/eigr/spawn#getting-started).\nBut you can also run your Elixir application in the traditional way as follows:\n\n```\nMIX_ENV=prod USER_FUNCTION_PORT=8092 PROXY_DATABASE_TYPE=mysql SPAWN_STATESTORE_KEY=3Jnb0hZiHIzHTOih7t2cTEPEpY98Tu1wvQkPfq/XwqE= iex --name spawn_a2@127.0.0.1 -S mix\n```\n\n> **_NOTE:_** For reasons of compatibility with our controller, it is necessary to configure your Spawn application using environment variables instead of the traditional Elixir configuration mechanism.\n\n> **_WARNING:_** This example uses the MySQL database as persistent storage for its actors. And it is also expected that you have previously created a database called eigr-functions-db in the MySQL instance.\n\nThe full example of this application can be found [here](https://github.com/eigr/spawn/tree/main/spawn_sdk/spawn_sdk_example).\n\nAnd links to other examples can be found in our github [readme page](https://github.com/eigr/spawn#examples).","ref":"SpawnSdk.html#module-running","title":"Running - SpawnSdk","type":"module"},{"doc":"To invoke Actors, use:\n\n```elixir\niex> SpawnSdk.invoke(\"joe\", system: \"spawn-system\", action: \"sum\", payload: %Io.Eigr.Spawn.Example.MyBusinessMessage{value: 1})\n{:ok, %Io.Eigr.Spawn.Example.MyBusinessMessage{value: 12}}\n```\n\nYou can invoke actor default functions like \"get\" to get its current state\n\n```elixir\nSpawnSdk.invoke(\"joe\", system: \"spawn-system\", action: \"get\")\n```\n\nSpawning Actors:\n\n```elixir\niex> SpawnSdk.spawn_actor(\"robert\", system: \"spawn-system\", actor: \"unamed_actor\")\n:ok\n```\n\nYou can also create Actors so that they are initialized from a certain revision number, that is, initialize actors from a specific point in time.\n\n```elixir\niex> SpawnSdk.spawn_actor(\"robert\", system: \"spawn-system\", actor: \"unamed_actor\", revision: 2)\n:ok\n```\n\nIn the above case the actor will be initialized with its state restored from the state as it was in revision 2 of its previous lifetime.\n\nInvoke Spawned Actors:\n\n```elixir\niex> SpawnSdk.invoke(\"robert\", system: \"spawn-system\", action: \"sum\", payload: %Io.Eigr.Spawn.Example.MyBusinessMessage{value: 1})\n{:ok, %Io.Eigr.Spawn.Example.MyBusinessMessage{value: 16}}\n```\n\nInvoke Actors in a lazy way without having to spawn them before:\n\n```elixir\niex> SpawnSdk.invoke(\"robert_lazy\", ref: SpawnSdkExample.Actors.UnamedActor, system: \"spawn-system\", action: \"sum\", payload: %Io.Eigr.Spawn.Example.MyBusinessMessage{value: 1})\n{:ok, %Io.Eigr.Spawn.Example.MyBusinessMessage{value: 1}}\n```\n\nInvoke Actors with a delay set in milliseconds:\n\n```elixir\niex> SpawnSdk.invoke(\"joe\", system: \"spawn-system\", action: \"ping\", delay: 5_000)\n{:ok, :async}\n```\n\nInvoke Actors scheduled to a specific DateTime:\n\n```elixir\niex> SpawnSdk.invoke(\"joe\", system: \"spawn-system\", action: \"ping\", scheduled_to: ~U[2023-01-01 00:32:00.145Z])\n{:ok, :async}\n```\n\nInvoke Pooled Actors:\n\n```elixir\niex> SpawnSdk.invoke(\"pooled_actor\", system: \"spawn-system\", action: \"ping\", pooled: true)\n{:ok, nil}\n```","ref":"SpawnSdk.html#module-client-api-examples","title":"Client API Examples - SpawnSdk","type":"module"},{"doc":"Invokes a function for a actor_name","ref":"SpawnSdk.html#invoke/2","title":"SpawnSdk.invoke/2","type":"function"},{"doc":"- `system` this is required\n- `ref` attribute attribute will always lookup to see if the referenced actor is already started or not.\n- `payload` attribute is optional.\n- `action` has default values that you can use to get current actor state\n  - get, get_state, Get, getState, GetState","ref":"SpawnSdk.html#invoke/2-opts","title":"Opts - SpawnSdk.invoke/2","type":"function"},{"doc":"```elixir\niex> SpawnSdk.invoke(\n  \"actor_name\",\n  ref: SpawnSdkExample.Actors.UnamedActor,\n  system: \"spawn-system\",\n  action: \"sum\", # \"sum\" or :sum\n  payload: %Io.Eigr.Spawn.Example.MyBusinessMessage{value: 5}\n)\n{:ok, %Io.Eigr.Spawn.Example.MyBusinessMessage{value: 5}}\n\niex> SpawnSdk.invoke(\"actor_name\", system: \"spawn-system\", action: \"get\")\n{:ok, %Io.Eigr.Spawn.Example.MyBusinessMessage{value: 5}}\n```","ref":"SpawnSdk.html#invoke/2-examples","title":"Examples - SpawnSdk.invoke/2","type":"function"},{"doc":"Spawns a Unamed actor\n\nA Unamed actor means that you can spawn dynamically the same actor for multiple different names.\nIt is analog to `DynamicSupervisor`","ref":"SpawnSdk.html#spawn_actor/2","title":"SpawnSdk.spawn_actor/2","type":"function"},{"doc":"- `system` this is required\n- `actor` which actor you will register first argument to","ref":"SpawnSdk.html#spawn_actor/2-opts","title":"Opts - SpawnSdk.spawn_actor/2","type":"function"},{"doc":"```elixir\niex> SpawnSdk.spawn_actor(\n  \"actor_name\",\n  system: \"spawn-system\",\n  actor: SpawnSdkExample.Actors.UnamedActor\n)\n```","ref":"SpawnSdk.html#spawn_actor/2-examples","title":"Examples - SpawnSdk.spawn_actor/2","type":"function"},{"doc":"Documentation for `Actor`.\n\nActor look like this:\n\n  defmodule MyActor do\n    use SpawnSdk.Actor,\n      name: \"joe\",\n      persistent: false,\n      state_type: Io.Eigr.Spawn.Example.MyState,\n      deactivate_timeout: 5_000,\n      snapshot_timeout: 2_000\n\n    require Logger\n    alias Io.Eigr.Spawn.Example.{MyState, MyBusinessMessage}\n\n    defact sum(%MyBusinessMessage{value: value} = data}, %Context{state: state} = ctx) do\n      Logger.info(\"Received Request...\")\n\n      new_value = (state.value || 0) + value\n\n      %Value{}\n      |> Value.of(%MyBusinessMessage{value: new_value}, %MyState{value: new_value})\n      |> Value.reply!()\n    end","ref":"SpawnSdk.Actor.html","title":"SpawnSdk.Actor","type":"behaviour"},{"doc":"Sends a assynchronous message to group of actors.\n\nExample:\n\n```\nalias SpawnSdk.Actor\n\nActor.ref(\"spawn-system\", \"joe\")\n|> Actor.cast(action: \"sum\", data: %MyData{value: 1})\n```","ref":"SpawnSdk.Actor.html#cast/2","title":"SpawnSdk.Actor.cast/2","type":"function"},{"doc":"","ref":"SpawnSdk.Actor.html#channel/2","title":"SpawnSdk.Actor.channel/2","type":"function"},{"doc":"","ref":"SpawnSdk.Actor.html#group/2","title":"SpawnSdk.Actor.group/2","type":"function"},{"doc":"","ref":"SpawnSdk.Actor.html#c:handle_action/2","title":"SpawnSdk.Actor.handle_action/2","type":"callback"},{"doc":"Sends a message to the actor and returns the result.\n\nExample:\n\n```\nalias SpawnSdk.Actor\n\nActor.ref(\"spawn-system\", \"joe\")\n|> Actor.invoke(action: \"sum\", data: %MyData{value: 1})\n```","ref":"SpawnSdk.Actor.html#invoke/2","title":"SpawnSdk.Actor.invoke/2","type":"function"},{"doc":"Invokes a group of actors and returns all results.\n\nExample:\n\n```\nalias SpawnSdk.Actor\n\nmy_data = %MyData{value: 1}\n\nActor.ref(\"erlang-system\", \"joe\", action: :sum, data: my_data)\n|> Actor.to_group(\"erlang-system\", \"robert\", action: :sum, data: my_data)\n|> Actor.to_group(\"erlang-system\", \"mike\", action: :sum, data: my_data)\n|> Actor.to_group(\"eigr-system\", \"adriano\", action: :sum, data: my_data)\n|> Actor.to_group(\"eigr-system\", \"marcel\", action: :sum, data: my_data)\n|> Actor.to_group(\"spawn-elixir-system\", \"elias\"\", action: \"calc\", data: my_data)\n|> Actor.multi()\n```","ref":"SpawnSdk.Actor.html#multi/1","title":"SpawnSdk.Actor.multi/1","type":"function"},{"doc":"","ref":"SpawnSdk.Actor.html#pub/2","title":"SpawnSdk.Actor.pub/2","type":"function"},{"doc":"Creates a reference to an actor so that it can be invoked.\n\nThe first argument is ActorSystem name.\nThe second argument is a Actor name.\n\nThe third argument is a keyword list of options:\n\n* `spawn` - a boolean indicating whether the actor should be generated or not. Default is **false**.\n\nExample:\n\n```\niex(spawn_a@127.0.0.1)1> SpawnSdk.Actor.ref(\"spawn-system\", \"joe\")\n%SpawnSdk.ActorRef{system: \"spawn-system\", name: \"joe\", opts: []}\n```\n\nTo invoke an actor using the obtained reference you could do:\n\n```\nalias SpawnSdk.Actor\n\nmy_data = %MyData{value: 1}\n\nActor.ref(\"spawn-system\", \"joe\")\n|> Actor.invoke(action: \"sum\", data: my_data)\n```","ref":"SpawnSdk.Actor.html#ref/3","title":"SpawnSdk.Actor.ref/3","type":"function"},{"doc":"Spawn an actor and return its reference.\n\nThe first argument is ActorSystem name.\nThe second argument is a Actor name.\nThe third argument is a parent actor name.\n\nExample:\n\n```\niex(spawn_a@127.0.0.1)1> SpawnSdk.Actor.spawn(\"spawn-system\", \"joe\", \"unamed_actor\")\n%SpawnSdk.ActorRef{system: \"spawn-system\", name: \"joe\", opts: []}\n```\n\nTo invoke an actor using the obtained reference you could do:\n\n```\nalias SpawnSdk.Actor\n\nmy_data = %MyData{value: 1}\n\nActor.spawn(\"spawn-system\", \"joe\", \"unamed_actor\")\n|> Actor.invoke(action: \"sum\", data: my_data)\n```","ref":"SpawnSdk.Actor.html#spawn/4","title":"SpawnSdk.Actor.spawn/4","type":"function"},{"doc":"Creates a group of actor reference so that it can be invoked\n\nExample:\n\n```\nalias SpawnSdk.Actor\n\nmy_data = %MyData{value: 1}\n\nActor.ref(\"erlang-system\", \"joe\", action: :sum, data: my_data)\n|> Actor.to_group(\"erlang-system\", \"robert\", action: :sum, data: my_data)\n|> Actor.multi()\n```","ref":"SpawnSdk.Actor.html#to_group/2","title":"SpawnSdk.Actor.to_group/2","type":"function"},{"doc":"Creates a group of actor reference so that it can be invoked\n\nExample:\n\n```\nalias SpawnSdk.Actor\n\nmy_data = %MyData{value: 1}\n\nActor.ref(\"erlang-system\", \"joe\")\n|> Actor.to_group(\"erlang-system\", \"robert\")\n|> Actor.to_group(\"erlang-system\", \"mike\")\n|> Actor.to_group(\"eigr-system\", \"adriano\")\n|> Actor.to_group(\"eigr-system\", \"marcel\")\n|> Actor.to_group(\"spawn-elixir-system\", \"elias\")\n|> Actor.multi(action: :sum, data: my_data)\n```","ref":"SpawnSdk.Actor.html#to_group/4","title":"SpawnSdk.Actor.to_group/4","type":"function"},{"doc":"","ref":"SpawnSdk.Actor.html#t:action/0","title":"SpawnSdk.Actor.action/0","type":"type"},{"doc":"","ref":"SpawnSdk.Actor.html#t:actor/0","title":"SpawnSdk.Actor.actor/0","type":"type"},{"doc":"","ref":"SpawnSdk.Actor.html#t:context/0","title":"SpawnSdk.Actor.context/0","type":"type"},{"doc":"","ref":"SpawnSdk.Actor.html#t:data/0","title":"SpawnSdk.Actor.data/0","type":"type"},{"doc":"","ref":"SpawnSdk.Actor.html#t:error/0","title":"SpawnSdk.Actor.error/0","type":"type"},{"doc":"","ref":"SpawnSdk.Actor.html#t:group/0","title":"SpawnSdk.Actor.group/0","type":"type"},{"doc":"","ref":"SpawnSdk.Actor.html#t:opts/0","title":"SpawnSdk.Actor.opts/0","type":"type"},{"doc":"","ref":"SpawnSdk.Actor.html#t:parent/0","title":"SpawnSdk.Actor.parent/0","type":"type"},{"doc":"","ref":"SpawnSdk.Actor.html#t:system/0","title":"SpawnSdk.Actor.system/0","type":"type"},{"doc":"","ref":"SpawnSdk.Actor.html#t:value/0","title":"SpawnSdk.Actor.value/0","type":"type"},{"doc":"","ref":"SpawnSdk.ActorChannel.html","title":"SpawnSdk.ActorChannel","type":"module"},{"doc":"","ref":"SpawnSdk.ActorChannel.html#t:t/0","title":"SpawnSdk.ActorChannel.t/0","type":"type"},{"doc":"","ref":"SpawnSdk.ActorGroupRef.html","title":"SpawnSdk.ActorGroupRef","type":"module"},{"doc":"","ref":"SpawnSdk.ActorGroupRef.html#t:t/0","title":"SpawnSdk.ActorGroupRef.t/0","type":"type"},{"doc":"","ref":"SpawnSdk.ActorRef.html","title":"SpawnSdk.ActorRef","type":"module"},{"doc":"","ref":"SpawnSdk.ActorRef.html#t:t/0","title":"SpawnSdk.ActorRef.t/0","type":"type"},{"doc":"`Subscriber` is a helper module to subscribe Phoenix.PubSub channels","ref":"SpawnSdk.Channel.Subscriber.html","title":"SpawnSdk.Channel.Subscriber","type":"module"},{"doc":"","ref":"SpawnSdk.Channel.Subscriber.html#subscribe/2","title":"SpawnSdk.Channel.Subscriber.subscribe/2","type":"function"},{"doc":"","ref":"SpawnSdk.Channel.Subscriber.html#unsubscribe/1","title":"SpawnSdk.Channel.Subscriber.unsubscribe/1","type":"function"},{"doc":"","ref":"SpawnSdk.Channel.Subscriber.html#t:channel/0","title":"SpawnSdk.Channel.Subscriber.channel/0","type":"type"},{"doc":"","ref":"SpawnSdk.Channel.Subscriber.html#t:opts/0","title":"SpawnSdk.Channel.Subscriber.opts/0","type":"type"},{"doc":"The context is responsible for sending the State information as well as its metadata\nto the Actor and the Proxy and vice versa.","ref":"SpawnSdk.Context.html","title":"SpawnSdk.Context","type":"module"},{"doc":"","ref":"SpawnSdk.Context.html#t:t/0","title":"SpawnSdk.Context.t/0","type":"type"},{"doc":"Define actions like a Elixir functions\n\n#","ref":"SpawnSdk.Defact.html","title":"SpawnSdk.Defact","type":"module"},{"doc":"[\n  {action_name, %{timer: 10_000}},\n  {action_name2, %{timer: nil}}\n]","ref":"SpawnSdk.Defact.html#module-internal-defact_exports-metadata-saved-as","title":"Internal :defact_exports metadata saved as - SpawnSdk.Defact","type":"module"},{"doc":"","ref":"SpawnSdk.Defact.html#defact/2","title":"SpawnSdk.Defact.defact/2","type":"macro"},{"doc":"","ref":"SpawnSdk.Flow.html","title":"SpawnSdk.Flow","type":"module"},{"doc":"Actors can also send messages to a group of actors at once as an action callback. This we call Broadcast.\n\n#","ref":"SpawnSdk.Flow.Broadcast.html","title":"SpawnSdk.Flow.Broadcast","type":"module"},{"doc":"defmodule Fleet.Actors.Driver do\n  use SpawnSdk.Actor,\n    kind: :abstract,\n    state_type: Fleet.Domain.Driver\n\n  alias Fleet.Domain.{\n    Driver,\n    OfferRequest,\n    OfferResponse,\n    Point\n  }\n\n  require Logger\n\n  @brain_actor_channel \"fleet-controllers\"\n\n  defact update_position(%Point{} = position, %Context{state: %Driver{id: name} = driver} = ctx) do\n    driver_state = %Driver{driver | position: position}\n\n    %Value{}\n    |> Value.of(driver_state, driver_state)\n    |> Value.broadcast(\n      Broadcast.to(\n        @brain_actor_channel,\n        driver_state\n      )\n    )\n    |> Value.reply!()\n  end\nend\n\ndefmodule Fleet.Actors.FleetControllersActor do\n  use SpawnSdk.Actor,\n    kind: :unamed,\n    channels: [\n      {\"fleet.controllers.topic\", \"update_position_receive\"}\n    ] # or just [\"fleet.controllers.topic\"] and it will forward to a action called receive\n\n  alias Fleet.Domain.Point\n\n  defact update_position_receive(%Point{} = position, _ctx) do\n    Logger.info(\n      \"Received Update Position Event. Position: [{inspect(position)}]\"\n    )\n\n    Value.of()\n  end\nend\n\nIn the case above, every time an Actor \"driver\" executes the update_position action\nit will send a message to all the actors participating in the channel called \"fleet-controllers\".\n\nBroadcasts can also be performed outside the Spawn Actor system,\nusing the transport mechanism based on Phoenix.PubSub in memory or\nPhoenix.PubSub over Nats Broker.","ref":"SpawnSdk.Flow.Broadcast.html#module-example-using-elixir-sdk","title":"Example using Elixir SDK: - SpawnSdk.Flow.Broadcast","type":"module"},{"doc":"","ref":"SpawnSdk.Flow.Broadcast.html#to/2","title":"SpawnSdk.Flow.Broadcast.to/2","type":"function"},{"doc":"","ref":"SpawnSdk.Flow.Broadcast.html#t:channel/0","title":"SpawnSdk.Flow.Broadcast.channel/0","type":"type"},{"doc":"","ref":"SpawnSdk.Flow.Broadcast.html#t:payload/0","title":"SpawnSdk.Flow.Broadcast.payload/0","type":"type"},{"doc":"","ref":"SpawnSdk.Flow.Broadcast.html#t:t/0","title":"SpawnSdk.Flow.Broadcast.t/0","type":"type"},{"doc":"Forward allows the Actor to delegate processing of the incoming message to another Actor.\nThis is done as part of the actor's response flow.\nForwards are detached from the Actor that received the input, that is,\nwhen you forward a message to another actor, the actor that performs the forwarding is free\nto process another message and the actor that is receiving the forwarding will respond\nto the original caller.","ref":"SpawnSdk.Flow.Forward.html","title":"SpawnSdk.Flow.Forward","type":"module"},{"doc":"","ref":"SpawnSdk.Flow.Forward.html#to/2","title":"SpawnSdk.Flow.Forward.to/2","type":"function"},{"doc":"","ref":"SpawnSdk.Flow.Forward.html#t:action/0","title":"SpawnSdk.Flow.Forward.action/0","type":"type"},{"doc":"","ref":"SpawnSdk.Flow.Forward.html#t:actor_name/0","title":"SpawnSdk.Flow.Forward.actor_name/0","type":"type"},{"doc":"","ref":"SpawnSdk.Flow.Forward.html#t:t/0","title":"SpawnSdk.Flow.Forward.t/0","type":"type"},{"doc":"Pipe allows the Actor to send its output message directly to another Actor,\nwhere the Actor that receives the Pipe will be responsible for following the flow from then on.\nThis is done as part of the actor's response flow.\nPipes are detached from the Actor that received the input, that is,\nwhen you forward a message to another actor through a Pipe,\nthe actor that performs the Pipe is free to process another message\nand the actor that is receiving the Pipe is the one who will respond to the original caller.","ref":"SpawnSdk.Flow.Pipe.html","title":"SpawnSdk.Flow.Pipe","type":"module"},{"doc":"","ref":"SpawnSdk.Flow.Pipe.html#to/2","title":"SpawnSdk.Flow.Pipe.to/2","type":"function"},{"doc":"","ref":"SpawnSdk.Flow.Pipe.html#t:action/0","title":"SpawnSdk.Flow.Pipe.action/0","type":"type"},{"doc":"","ref":"SpawnSdk.Flow.Pipe.html#t:actor_name/0","title":"SpawnSdk.Flow.Pipe.actor_name/0","type":"type"},{"doc":"","ref":"SpawnSdk.Flow.Pipe.html#t:t/0","title":"SpawnSdk.Flow.Pipe.t/0","type":"type"},{"doc":"Actors can also emit side effects to other Actors as part of their response.\nSide effects do not interfere with an actor's request-response flow.\nThey will \"always\" be processed asynchronously and any response sent back from the Actor\nreceiving the effect will be ignored by the effector.","ref":"SpawnSdk.Flow.SideEffect.html","title":"SpawnSdk.Flow.SideEffect","type":"module"},{"doc":"","ref":"SpawnSdk.Flow.SideEffect.html#effect/5","title":"SpawnSdk.Flow.SideEffect.effect/5","type":"function"},{"doc":"","ref":"SpawnSdk.Flow.SideEffect.html#of/0","title":"SpawnSdk.Flow.SideEffect.of/0","type":"function"},{"doc":"","ref":"SpawnSdk.Flow.SideEffect.html#to/4","title":"SpawnSdk.Flow.SideEffect.to/4","type":"function"},{"doc":"","ref":"SpawnSdk.Flow.SideEffect.html#t:action/0","title":"SpawnSdk.Flow.SideEffect.action/0","type":"type"},{"doc":"","ref":"SpawnSdk.Flow.SideEffect.html#t:actor_name/0","title":"SpawnSdk.Flow.SideEffect.actor_name/0","type":"type"},{"doc":"","ref":"SpawnSdk.Flow.SideEffect.html#t:payload/0","title":"SpawnSdk.Flow.SideEffect.payload/0","type":"type"},{"doc":"","ref":"SpawnSdk.Flow.SideEffect.html#t:t/0","title":"SpawnSdk.Flow.SideEffect.t/0","type":"type"},{"doc":"Implements the communication protocol between Elixir SDK and Sidecar.","ref":"SpawnSdk.Interface.html","title":"SpawnSdk.Interface","type":"module"},{"doc":"`System` defines the general behavior of the Spawn actor system.\nIt is through System implementations that the user can register, invoke,\nand perform other activities with their Actors.","ref":"SpawnSdk.System.html","title":"SpawnSdk.System","type":"behaviour"},{"doc":"","ref":"SpawnSdk.System.html#c:invoke/2","title":"SpawnSdk.System.invoke/2","type":"callback"},{"doc":"","ref":"SpawnSdk.System.html#c:register/2","title":"SpawnSdk.System.register/2","type":"callback"},{"doc":"","ref":"SpawnSdk.System.html#c:spawn_actor/2","title":"SpawnSdk.System.spawn_actor/2","type":"callback"},{"doc":"","ref":"SpawnSdk.System.html#t:action/0","title":"SpawnSdk.System.action/0","type":"type"},{"doc":"","ref":"SpawnSdk.System.html#t:actor_mod/0","title":"SpawnSdk.System.actor_mod/0","type":"type"},{"doc":"","ref":"SpawnSdk.System.html#t:actor_name/0","title":"SpawnSdk.System.actor_name/0","type":"type"},{"doc":"","ref":"SpawnSdk.System.html#t:actors/0","title":"SpawnSdk.System.actors/0","type":"type"},{"doc":"","ref":"SpawnSdk.System.html#t:invoke_opts/0","title":"SpawnSdk.System.invoke_opts/0","type":"type"},{"doc":"","ref":"SpawnSdk.System.html#t:payload/0","title":"SpawnSdk.System.payload/0","type":"type"},{"doc":"","ref":"SpawnSdk.System.html#t:spawn_actor_opts/0","title":"SpawnSdk.System.spawn_actor_opts/0","type":"type"},{"doc":"","ref":"SpawnSdk.System.html#t:system/0","title":"SpawnSdk.System.system/0","type":"type"},{"doc":"`SpawnSystem`","ref":"SpawnSdk.System.SpawnSystem.html","title":"SpawnSdk.System.SpawnSystem","type":"module"},{"doc":"","ref":"SpawnSdk.System.SpawnSystem.html#call/3","title":"SpawnSdk.System.SpawnSystem.call/3","type":"function"},{"doc":"","ref":"SpawnSdk.System.SpawnSystem.html#decode_kind/1","title":"SpawnSdk.System.SpawnSystem.decode_kind/1","type":"function"},{"doc":"","ref":"SpawnSdk.System.SpawnSystem.html#delete_cached_actors/2","title":"SpawnSdk.System.SpawnSystem.delete_cached_actors/2","type":"function"},{"doc":"hey","ref":"SpawnSdk.System.SpawnSystem.html#invoke/2","title":"SpawnSdk.System.SpawnSystem.invoke/2","type":"function"},{"doc":"","ref":"SpawnSdk.System.SpawnSystem.html#merge_cache_actors/2","title":"SpawnSdk.System.SpawnSystem.merge_cache_actors/2","type":"function"},{"doc":"Documentation for `System.Supervisor`.\n\nStart Supervisor tree like this:\n\n  defmodule MyApp.Application do\n    use Application\n\n    @impl true\n    def start(_type, _args) do\n      children = [\n        {\n          SpawnSdk.System.Supervisor,\n          system: \"my-system\", actors: [MyActor]\n        }\n      ]\n\n      opts = [strategy: :one_for_one, name: MyApp.Supervisor]\n      Supervisor.start_link(children, opts)\n  end","ref":"SpawnSdk.System.Supervisor.html","title":"SpawnSdk.System.Supervisor","type":"module"},{"doc":"Returns a specification to start this module under a supervisor.\n\nSee `Supervisor`.","ref":"SpawnSdk.System.Supervisor.html#child_spec/1","title":"SpawnSdk.System.Supervisor.child_spec/1","type":"function"},{"doc":"","ref":"SpawnSdk.System.Supervisor.html#start_link/1","title":"SpawnSdk.System.Supervisor.start_link/1","type":"function"},{"doc":"`Value` is a declarative DSL that provides the Domain Driven aspect of the Spawn technology.\nIt is through a Value that the user can configure the proxy to carry out\nthe infrastructure tasks and from which it provides all the Worflows.","ref":"SpawnSdk.Value.html","title":"SpawnSdk.Value","type":"module"},{"doc":"","ref":"SpawnSdk.Value.html#broadcast/2","title":"SpawnSdk.Value.broadcast/2","type":"function"},{"doc":"","ref":"SpawnSdk.Value.html#effect/2","title":"SpawnSdk.Value.effect/2","type":"function"},{"doc":"","ref":"SpawnSdk.Value.html#effects/2","title":"SpawnSdk.Value.effects/2","type":"function"},{"doc":"","ref":"SpawnSdk.Value.html#forward/2","title":"SpawnSdk.Value.forward/2","type":"function"},{"doc":"","ref":"SpawnSdk.Value.html#noreply!/2","title":"SpawnSdk.Value.noreply!/2","type":"function"},{"doc":"","ref":"SpawnSdk.Value.html#noreply_state!/1","title":"SpawnSdk.Value.noreply_state!/1","type":"function"},{"doc":"","ref":"SpawnSdk.Value.html#of/0","title":"SpawnSdk.Value.of/0","type":"function"},{"doc":"","ref":"SpawnSdk.Value.html#of/2","title":"SpawnSdk.Value.of/2","type":"function"},{"doc":"","ref":"SpawnSdk.Value.html#of/3","title":"SpawnSdk.Value.of/3","type":"function"},{"doc":"","ref":"SpawnSdk.Value.html#pipe/2","title":"SpawnSdk.Value.pipe/2","type":"function"},{"doc":"","ref":"SpawnSdk.Value.html#reply!/1","title":"SpawnSdk.Value.reply!/1","type":"function"},{"doc":"","ref":"SpawnSdk.Value.html#response/2","title":"SpawnSdk.Value.response/2","type":"function"},{"doc":"","ref":"SpawnSdk.Value.html#state/2","title":"SpawnSdk.Value.state/2","type":"function"},{"doc":"","ref":"SpawnSdk.Value.html#tags/2","title":"SpawnSdk.Value.tags/2","type":"function"},{"doc":"","ref":"SpawnSdk.Value.html#value/2","title":"SpawnSdk.Value.value/2","type":"function"},{"doc":"","ref":"SpawnSdk.Value.html#void/0","title":"SpawnSdk.Value.void/0","type":"function"},{"doc":"","ref":"SpawnSdk.Value.html#void/1","title":"SpawnSdk.Value.void/1","type":"function"},{"doc":"","ref":"SpawnSdk.Value.html#t:broadcast/0","title":"SpawnSdk.Value.broadcast/0","type":"type"},{"doc":"","ref":"SpawnSdk.Value.html#t:effects/0","title":"SpawnSdk.Value.effects/0","type":"type"},{"doc":"","ref":"SpawnSdk.Value.html#t:forward/0","title":"SpawnSdk.Value.forward/0","type":"type"},{"doc":"","ref":"SpawnSdk.Value.html#t:new_state/0","title":"SpawnSdk.Value.new_state/0","type":"type"},{"doc":"","ref":"SpawnSdk.Value.html#t:pipe/0","title":"SpawnSdk.Value.pipe/0","type":"type"},{"doc":"","ref":"SpawnSdk.Value.html#t:response/0","title":"SpawnSdk.Value.response/0","type":"type"},{"doc":"","ref":"SpawnSdk.Value.html#t:t/0","title":"SpawnSdk.Value.t/0","type":"type"},{"doc":"","ref":"SpawnSdk.Value.html#t:tags/0","title":"SpawnSdk.Value.tags/0","type":"type"},{"doc":"","ref":"SpawnSdk.Value.html#t:value/0","title":"SpawnSdk.Value.value/0","type":"type"},{"doc":"# Spawn Elixir SDK\n\n \n\nSpawn Elixir SDK is the support library for the Spawn Actors System.\nSpawn is a Stateful Serverless Platform for providing the multi-language Actor Model. For a broader understanding of Spawn, please consult its official [repository](https://github.com/eigr/spawn).\n\nThe advantage of the Elixir SDK over other SDKs is in Elixir's native ability to connect directly to an Erlang network. For this reason, the Elixir SDK allows any valid Elixir application to be part of a Spawn network without needing a sidecar attached.","ref":"readme.html","title":"Spawn Elixir SDK","type":"extras"},{"doc":"[Available in Hex](https://hex.pm/packages/spawn_sdk), the package can be installed\nby adding `spawn_sdk` and `spawn_statestores_*` to your list of dependencies in `mix.exs`:\n\n```elixir\ndef deps do\n  [\n    {:spawn_sdk, \"~> 1.0.1\"},\n\n    # You can uncomment one of those dependencies if you are going to use Persistent Actors\n    #{:spawn_statestores_mariadb, \"~> 1.0.1\"},\n    #{:spawn_statestores_mysql, \"~> 1.0.1\"},\n    #{:spawn_statestores_postgres, \"~> 1.0.1\"},\n    #{:spawn_statestores_mssql, \"~> 1.0.1\"},\n    #{:spawn_statestores_cockroachdb, \"~> 1.0.1\"},\n    #{:spawn_statestores_sqlite, \"~> 1.0.1\"},\n  ]\nend\n```\n\n#","ref":"readme.html#installation","title":"Installation - Spawn Elixir SDK","type":"extras"},{"doc":"Following the steps below, you will have a valid Elixir application to use in a Spawn cluster. However, you will still need to generate a container image with your application to use it together with the [Spawn Operator for Kubernetes](https://github.com/eigr/spawn/tree/main/spawn_operator/spawn_operator).","ref":"readme.html#deploy","title":"Deploy - Spawn Elixir SDK","type":"extras"},{"doc":"After creating an Elixir application project, create the protobuf files for your business domain.\nIt is common practice to do this under the **priv/** folder. Let's demonstrate an example:\n\n```protobuf\nsyntax = \"proto3\";\n\npackage io.eigr.spawn.example;\n\nmessage MyState {\n  int32 value = 1;\n}\n\nmessage MyBusinessMessage {\n  int32 value = 1;\n}\n```\n\nIt is important to try to separate the type of message that must be stored as the actors' state from the type of messages\nthat will be exchanged between their actors' operations calls. In other words, the Actor's internal state is also represented\nas a protobuf type, and it is a good practice to separate these types of messages from the others in its business domain.\n\nIn the above case `MyState` is the type protobuf that represents the state of the Actor that we will create later\nwhile `MyBusiness` is the type of message that we will send and receive from this Actor.\n\nNow that we have defined our input and output types as Protobuf types we will need to compile these files to generate their respective Elixir modules. An example of how to do this can be found [here](https://github.com/eigr/spawn/blob/main/spawn_sdk/spawn_sdk_example/compile-example-pb.sh)\n\n> **_NOTE:_** You need to have installed the elixir plugin for protoc. More information on how to obtain and install the necessary tools can be found here [here](https://github.com/elixir-protobuf/protobuf#usage)\n\nNow that the protobuf types have been created we can proceed with the code. Example definition of an Actor.","ref":"readme.html#how-to-use","title":"How to use - Spawn Elixir SDK","type":"extras"},{"doc":"In this example we are creating an actor in a Named way, that is, it is a known actor at compile time.\n\n```elixir\ndefmodule SpawnSdkExample.Actors.MyActor do\n  use SpawnSdk.Actor,\n    name: \"jose\", # Default is Full Qualified Module name a.k.a __MODULE__\n    kind: :named, # Default is already :named. Valid are :named | :unamed | :pooled\n    stateful: true, # Default is already true\n    state_type: Io.Eigr.Spawn.Example.MyState, # or :json if you don't care about protobuf types\n    deactivate_timeout: 30_000,\n    snapshot_timeout: 2_000\n\n  require Logger\n\n  alias Io.Eigr.Spawn.Example.{MyState, MyBusinessMessage}\n\n  defact init(%Context{state: state} = ctx) do\n    Logger.info(\"[joe] Received InitRequest. Context: #{inspect(ctx)}\")\n\n    Value.of()\n    |> Value.state(state)\n  end\n\n  defact sum(\n           %MyBusinessMessage{value: value} = data,\n           %Context{state: state} = ctx\n         ) do\n    Logger.info(\"Received Request: #{inspect(data)}. Context: #{inspect(ctx)}\")\n\n    new_value = if is_nil(state), do: value, else: (state.value || 0) + value\n\n    Value.of(%MyBusinessMessage{value: new_value}, %MyState{value: new_value})\n  end\nend\n\n```\n\nWe declare two actions that the Actor can do. An initialization action that will be called every time an Actor instance is created and an action that will be responsible for performing a simple sum.\n\nNote Keep in mind that any Action that has the names present in the list below will behave as an initialization Action and will be called when the Actor is started (if there is more than one Action with one of these names, only one will be called).\n\nDefaults inicialization Action names: \"**init**\", \"**Init**\", \"**setup**\", \"**Setup**\"","ref":"readme.html#named-actors","title":"Named Actors - Spawn Elixir SDK","type":"extras"},{"doc":"We can also create Unnamed Dynamic/Lazy actors, that is, despite having its unamed behavior defined at compile time, a Lazy actor will only have a concrete instance when it is associated with an identifier/name at runtime. Below follows the same previous actor being defined as Unamed.\n\n```elixir\ndefmodule SpawnSdkExample.Actors.UnamedActor do\n  use SpawnSdk.Actor,\n    name: \"unamed_actor\",\n    kind: :unamed,\n    state_type: Io.Eigr.Spawn.Example.MyState\n\n  require Logger\n\n  alias Io.Eigr.Spawn.Example.{MyState, MyBusinessMessage}\n\n  defact sum(\n           %MyBusinessMessage{value: value} = data,\n           %Context{state: state} = ctx\n         ) do\n    Logger.info(\"Received Request: #{inspect(data)}. Context: #{inspect(ctx)}\")\n\n    new_value = if is_nil(state), do: value, else: (state.value || 0) + value\n\n    Value.of(%MyBusinessMessage{value: new_value}, %MyState{value: new_value})\n  end\nend\n```\n\nNotice that the only thing that has changed is the the kind of actor, in this case the kind is set to :unamed.\n\n> **_NOTE:_** Can Elixir programmers think in terms of Named vs Unamed actors as more or less known at startup vs dynamically supervised/registered? That is, defining your actors directly in the supervision tree or using a Dynamic Supervisor for that.","ref":"readme.html#unamed-actor","title":"Unamed Actor - Spawn Elixir SDK","type":"extras"},{"doc":"Sometimes we want a particular actor to be able to serve requests concurrently,\nhowever actors will always serve one request at a time using buffering mechanisms to receive requests in their mailbox and serve each request one by one.\nSo to get around this behaviour you can configure your Actor as a Pooled Actor, this way the system will generate a pool of actors to meet certain requests. See an example below:\n\n```elixir\ndefmodule SpawnSdkExample.Actors.PooledActor do\n  use SpawnSdk.Actor,\n    name: \"pooled_actor\",\n    kind: :pooled,\n    stateful: false\n\n  require Logger\n\n  defact ping(_data, %Context{} = ctx) do\n    Logger.info(\"Received Request. Context: #{inspect(ctx)}\")\n\n    Value.void()\n  end\nend\n```\n\n> **_NOTE:_** It is important to mention that Pooled Actors cannot be stateful because concurrent accesses for Stateful actors could generate undesired inconsistencies.\n\nTo invoke a pooled actor, it is necessary to pass the pooled: true option to the invoke function. See an example:\n\n```elixir\niex(spawn_a1@127.0.0.1)16> for _n   for _n   ]:[notice]:Activating Actor pooled_actor-5 with Parent pooled_actor-5 in Node :\"spawn_a1@127.0.0.1\". Persistence false.\n2022-11-25 11:21:53.759 [spawn_a1@127.0.0.1]:[pid=  ]:[debug]:Actor pooled_actor-5 reactivated. ActorRef PID: #PID \n2022-11-25 11:21:53.760 [spawn_a1@127.0.0.1]:[pid=  ]:[info]:Received Request. Context: %SpawnSdk.Context{state: nil, caller: nil, self: %Eigr.Functions.Protocol.Actors.ActorId{name: \"pooled_actor-5\", system: \"spawn-system\", parent: \"pooled_actor-5\", __unknown_fields__: []}, metadata: %{}}\n2022-11-25 11:21:53.763 [spawn_a1@127.0.0.1]:[pid=  ]:[notice]:Activating Actor pooled_actor-4 with Parent pooled_actor-4 in Node :\"spawn_a1@127.0.0.1\". Persistence false.\n2022-11-25 11:21:53.763 [spawn_a1@127.0.0.1]:[pid=  ]:[debug]:Actor pooled_actor-4 reactivated. ActorRef PID: #PID \n2022-11-25 11:21:53.765 [spawn_a1@127.0.0.1]:[pid=  ]:[info]:Received Request. Context: %SpawnSdk.Context{state: nil, caller: nil, self: %Eigr.Functions.Protocol.Actors.ActorId{name: \"pooled_actor-4\", system: \"spawn-system\", parent: \"pooled_actor-4\", __unknown_fields__: []}, metadata: %{}}\n2022-11-25 11:21:53.767 [spawn_a1@127.0.0.1]:[pid=  ]:[notice]:Activating Actor pooled_actor-7 with Parent pooled_actor-7 in Node :\"spawn_a1@127.0.0.1\". Persistence false.\n2022-11-25 11:21:53.767 [spawn_a1@127.0.0.1]:[pid=  ]:[debug]:Actor pooled_actor-7 reactivated. ActorRef PID: #PID \n2022-11-25 11:21:53.768 [spawn_a1@127.0.0.1]:[pid=  ]:[info]:Received Request. Context: %SpawnSdk.Context{state: nil, caller: nil, self: %Eigr.Functions.Protocol.Actors.ActorId{name: \"pooled_actor-7\", system: \"spawn-system\", parent: \"pooled_actor-7\", __unknown_fields__: []}, metadata: %{}}\n2022-11-25 11:21:53.770 [spawn_a1@127.0.0.1]:[pid=  ]:[notice]:Activating Actor pooled_actor-2 with Parent pooled_actor-2 in Node :\"spawn_a1@127.0.0.1\". Persistence false.\n2022-11-25 11:21:53.771 [spawn_a1@127.0.0.1]:[pid=  ]:[debug]:Actor pooled_actor-2 reactivated. ActorRef PID: #PID \n2022-11-25 11:21:53.772 [spawn_a1@127.0.0.1]:[pid=  ]:[info]:Received Request. Context: %SpawnSdk.Context{state: nil, caller: nil, self: %Eigr.Functions.Protocol.Actors.ActorId{name: \"pooled_actor-2\", system: \"spawn-system\", parent: \"pooled_actor-2\", __unknown_fields__: []}, metadata: %{}}\n2022-11-25 11:21:53.774 [spawn_a1@127.0.0.1]:[pid=  ]:[notice]:Activating Actor pooled_actor-8 with Parent pooled_actor-8 in Node :\"spawn_a1@127.0.0.1\". Persistence false.\n2022-11-25 11:21:53.775 [spawn_a1@127.0.0.1]:[pid=  ]:[debug]:Actor pooled_actor-8 reactivated. ActorRef PID: #PID \n2022-11-25 11:21:53.776 [spawn_a1@127.0.0.1]:[pid=  ]:[info]:Received Request. Context: %SpawnSdk.Context{state: nil, caller: nil, self: %Eigr.Functions.Protocol.Actors.ActorId{name: \"pooled_actor-8\", system: \"spawn-system\", parent: \"pooled_actor-8\", __unknown_fields__: []}, metadata: %{}}\n2022-11-25 11:21:53.777 [spawn_a1@127.0.0.1]:[pid=  ]:[notice]:Activating Actor pooled_actor-9 with Parent pooled_actor-9 in Node :\"spawn_a1@127.0.0.1\". Persistence false.\n2022-11-25 11:21:53.778 [spawn_a1@127.0.0.1]:[pid=  ]:[debug]:Actor pooled_actor-9 reactivated. ActorRef PID: #PID \n2022-11-25 11:21:53.778 [spawn_a1@127.0.0.1]:[pid=  ]:[info]:Received Request. Context: %SpawnSdk.Context{state: nil, caller: nil, self: %Eigr.Functions.Protocol.Actors.ActorId{name: \"pooled_actor-9\", system: \"spawn-system\", parent: \"pooled_actor-9\", __unknown_fields__: []}, metadata: %{}}\n2022-11-25 11:21:53.779 [spawn_a1@127.0.0.1]:[pid=  ]:[debug]:Lookup Actor pooled_actor. PID: #PID \n2022-11-25 11:21:53.780 [spawn_a1@127.0.0.1]:[pid=  ]:[info]:Received Request. Context: %SpawnSdk.Context{state: nil, caller: nil, self: %Eigr.Functions.Protocol.Actors.ActorId{name: \"pooled_actor-8\", system: \"spawn-system\", parent: \"pooled_actor-8\", __unknown_fields__: []}, metadata: %{}}\n2022-11-25 11:21:53.780 [spawn_a1@127.0.0.1]:[pid=  ]:[debug]:Lookup Actor pooled_actor. PID: #PID \n2022-11-25 11:21:53.781 [spawn_a1@127.0.0.1]:[pid=  ]:[info]:Received Request. Context: %SpawnSdk.Context{state: nil, caller: nil, self: %Eigr.Functions.Protocol.Actors.ActorId{name: \"pooled_actor-5\", system: \"spawn-system\", parent: \"pooled_actor-5\", __unknown_fields__: []}, metadata: %{}}\n2022-11-25 11:21:53.781 [spawn_a1@127.0.0.1]:[pid=  ]:[debug]:Lookup Actor pooled_actor. PID: #PID \n[\n  ok: nil,\n  ok: nil,\n  ok: nil,\n  ok: nil,\n  ok: nil,\n  ok: nil,\n  ok: nil,\n  ok: nil,\n  ok: nil\n]\niex(spawn_a1@127.0.0.1)17> 2022-11-25 11:21:53.782 [spawn_a1@127.0.0.1]:[pid=  ]:[info]:Received Request. Context: %SpawnSdk.Context{state: nil, caller: nil, self: %Eigr.Functions.Protocol.Actors.ActorId{name: \"pooled_actor-4\", system: \"spawn-system\", parent: \"pooled_actor-4\", __unknown_fields__: []}, metadata: %{}}\n2022-11-25 11:22:03.999 [spawn_a1@127.0.0.1]:[pid=  ]:[debug]:Deactivating actor pooled_actor-9 for timeout\n2022-11-25 11:22:04.226 [spawn_a1@127.0.0.1]:[pid=  ]:[debug]:Deactivating actor pooled_actor-5 for timeout\n2022-11-25 11:22:07.271 [spawn_a1@127.0.0.1]:[pid=  ]:[debug]:Deactivating actor pooled_actor-8 for timeout\n2022-11-25 11:22:10.167 [spawn_a1@127.0.0.1]:[pid=  ]:[debug]:Deactivating actor pooled_actor-7 for timeout\n2022-11-25 11:22:12.038 [spawn_a1@127.0.0.1]:[pid=  ]:[debug]:Deactivating actor pooled_actor-4 for timeout\n2022-11-25 11:22:12.613 [spawn_a1@127.0.0.1]:[pid=  ]:[debug]:Deactivating actor pooled_actor-2 for timeout\n```\n\nIt is possible to configure the minimum and maximum pool size by passing the min_pool_size and max_pool_size options. The actor below would create a pool of a maximum of 100 Actors:\n\n```elixir\ndefmodule SpawnSdkExample.Actors.PooledActor do\n  use SpawnSdk.Actor,\n    name: \"pooled_actor\",\n    kind: :pooled,\n    min_pool_size: 1,\n    max_pool_size: 10,\n    stateful: false\n\n  # Code omitted for brevity\nend\n```\n\n> **_ALERT:_** Keep in mind that a large pool size does not always mean better performance. In fact, smaller pool sizes sometimes often have a much greater effect in this regard. It is interesting that you experiment with different pool sizes until you find the one that best suits your workload.","ref":"readme.html#pooled-actors","title":"Pooled Actors - Spawn Elixir SDK","type":"extras"},{"doc":"Actors can also emit side effects to other Actors as part of their response. See an example:\n\n```elixir\ndefmodule SpawnSdkExample.Actors.UnamedActor do\n  use SpawnSdk.Actor,\n    kind: :unamed,\n    stateful: false,\n    state_type: Io.Eigr.Spawn.Example.MyState\n\n  require Logger\n\n  alias Io.Eigr.Spawn.Example.{MyState, MyBusinessMessage}\n\n  alias SpawnSdk.Flow.SideEffect\n\n  defact sum(%MyBusinessMessage{value: value} = data, %Context{state: state} = ctx) do\n    Logger.info(\"Received Request: #{inspect(data)}. Context: #{inspect(ctx)}\")\n\n    new_value = if is_nil(state), do: value, else: (state.value || 0) + value\n\n    result = %MyBusinessMessage{value: new_value}\n    new_state = %MyState{value: new_value}\n\n    Value.of()\n    |> Value.response(result)\n    |> Value.state(new_state)\n    |> Value.effects(\n      # This returns a list of side effects. In this case containing only one effect. However, multiple effects can be chained together,\n      # just by calling the effect function as shown here.\n      # The delay means that it will be fired asynchronously after 5000 milliseconds (5 seconds)\n      # If only one effect is desired, you can also choose to use the to/3 function together with Value.effect().\n      # Example: Values.effect(SideEffect.to(name, func, payload))\n      SideEffect.of()\n      |> SideEffect.effect(\"joe\", :sum, result, delay: 5_000, scheduled_to: ~U[2020-01-01 10:00:00.145Z])\n      # use delay or scheduled_to, not both\n    )\n  end\nend\n\n```\n\nIn the example above we see that the Actor joe will receive a request as a side effect from the Actor who issued this effect.\n\nSide effects do not interfere with an actor's request-response flow. They will \"always\" be processed asynchronously and any response sent back from the Actor receiving the effect will be ignored by the effector.","ref":"readme.html#side-effects","title":"Side Effects - Spawn Elixir SDK","type":"extras"},{"doc":"Actors can also route some actions to other actors as part of their response. See an example:\n\n```elixir\ndefmodule SpawnSdkExample.Actors.ForwardPipeActor do\n  use SpawnSdk.Actor,\n    name: \"pipeforward\",\n    kind: :named,\n    stateful: false\n\n  require Logger\n\n  alias Io.Eigr.Spawn.Example.MyBusinessMessage\n\n  defact forward_example(%MyBusinessMessage{} = msg, _ctx) do\n    Logger.info(\"Received request with #{msg.value}\")\n\n    Value.of()\n    |> Value.forward(\n      Forward.to(\"second_actor\", \"sum_plus_one\")\n    )\n    |> Value.void()\n  end\n\n  defact pipe_example(%MyBusinessMessage{} = msg, _ctx) do\n    Logger.info(\"Received request with #{msg.value}\")\n\n    Value.of()\n    |> Value.response(MyBusinessMessage.new(value: 999))\n    |> Value.pipe(\n      Pipe.to(\"second_actor\", \"sum_plus_one\")\n    )\n    |> Value.void()\n  end\nend\n\ndefmodule SpawnSdkExample.Actors.SecondActorExample do\n  use SpawnSdk.Actor,\n    name: \"second_actor\",\n    stateful: false\n\n  require Logger\n\n  alias Io.Eigr.Spawn.Example.MyBusinessMessage\n\n  defact sum_plus_one(%MyBusinessMessage{} = msg, _ctx) do\n    Logger.info(\"Received request with #{msg.value}\")\n\n    Value.of()\n    |> Value.response(MyBusinessMessage.new(value: msg.value + 1))\n    |> Value.void()\n  end\nend\n\n```\n\nWe are returning void in both examples so we dont care about what is being stored in the actor state.\n\nIn the case above, every time you call the `forward_example` the second_actor's `sum_plus_one` function will receive the value forwarded originally in the invocation as its input. The end result will be:\n\n```elixir\niex> SpawnSdk.invoke(\"pipeforward\", system: \"spawn-system\", action: \"forward_example\", payload: %Io.Eigr.Spawn.Example.MyBusinessMessage{value: 1})\n{:ok, %Io.Eigr.Spawn.Example.MyBusinessMessage{value: 2}}\n```\n\nFor the Pipe example, the the second_actor's `sum_plus_one` function will always receive `%MyBusinessMessage{value: 999}` due to getting the value from the previous specification in the `pipe_example` action, the end result will be:\n\n```elixir\niex> SpawnSdk.invoke(\"pipeforward\", system: \"spawn-system\", action: \"pipe_example\", payload: %Io.Eigr.Spawn.Example.MyBusinessMessage{value: 1})\n{:ok, %Io.Eigr.Spawn.Example.MyBusinessMessage{value: 1000}}\n```","ref":"readme.html#pipe-and-forward","title":"Pipe and Forward - Spawn Elixir SDK","type":"extras"},{"doc":"Actors can also send messages to a group of actors at once as an action callback. See the example below:\n\n```elixir\ndefmodule Fleet.Actors.Driver do\n  use SpawnSdk.Actor,\n    kind: :unamed,\n    state_type: Fleet.Domain.Driver\n\n  alias Fleet.Domain.{\n    Driver,\n    OfferRequest,\n    OfferResponse,\n    Point\n  }\n\n  require Logger\n\n  @brain_actor_channel \"fleet.controllers.topic\"\n\n  defact update_position(%Point{} = position, %Context{state: %Driver{id: name} = driver} = ctx) do\n    Logger.info(\n      \"Received Update Position Event. Position: [{inspect(position)}]. Context: #{inspect(ctx)}\"\n    )\n\n    driver_state = %Driver{driver | position: position}\n\n    %Value{}\n    |> Value.of(driver_state, driver_state)\n    |> Value.broadcast(\n      Broadcast.to(\n        @brain_actor_channel,\n        driver_state\n      )\n    )\n  end\nend\n\ndefmodule Fleet.Actors.FleetControllersActor do\n  use SpawnSdk.Actor,\n    kind: :unamed,\n    channels: [\n      {\"fleet.controllers.topic\", \"update_position_receive\"}\n    ] # or just [\"fleet.controllers.topic\"] and it will forward to a action called receive\n\n  alias Fleet.Domain.Point\n\n  defact update_position_receive(%Point{} = position, _ctx) do\n    Logger.info(\n      \"Driver [#{name}] Received Update Position Event. Position: [#{inspect(position)}]\"\n    )\n\n    Value.of()\n  end\nend\n```\n\nIn the case above, every time an Actor \"driver\" executes the update_position action it will send a message to all the actors participating in the channel called \"fleet-controllers\".\n\n#","ref":"readme.html#broadcast","title":"Broadcast - Spawn Elixir SDK","type":"extras"},{"doc":"Sometimes you may want to send events out of ActorSystem using Phoenix.PubSub.\nOne way to do this is to take advantage of the same Broadcast infrastructure that Spawn offers you but indicating an external channel. Below is an example:\n\n1. Create a Listener to receive the events using the `SpawnSdk.Channel.Subscriber` helper module.\n\n```elixir\ndefmodule SpawnSdkExample.Subscriber do\n  @moduledoc \"\"\"\n  This module exemplifies how to listen for pubsub events that were emitted by actors but that will be treated not by actors but as normal pubsub events.\n  This is particularly useful for integrations between Spawn and Phoenix LiveView.\n  \"\"\"\n  use GenServer\n  require logger\n\n  alias SpawnSdk.Channel.Subscriber\n\n  @impl true\n  define init(state) do\n    Subscriber.subscribe(\"external.channel\")\n    {:ok, state}\n  end\n\n  @impl true\n  def handle_info({:receive, payload}, state) do\n    Logger.info(\"Received pubsub event #{inspect(payload)}\")\n    {:noreply, state}\n  end\n\n  def start_link(args) do\n    GenServer.start_link(__MODULE__, args)\n  end\nend\n```\n\nYou need to match using the {:receive, payload} tuple in your handle_info.\n\n> **_NOTE:_** By default SpawnSdk.Channel.Subscriber will subscribe to pubsub using the atom :actor_channel as an argument.\n\n    If you need to change this, just configure your configuration as follows:\n\n**_config.exs_**\n\n```elixir\nconfig :spawn,\n   pubsub_group: :your_channel_group_here\n```\n\n2. SpawnSdk.System.Supervisor.\n\n```elixir\ndefmodule SpawnSdkExample.Application do\n  @moduledoc false\n  use Application\n\n  @impl true\n  def start(_type, _args) do\n    children = [\n      {\n        SpawnSdk.System.Supervisor,\n        system: \"spawn-system\",\n        actors: [\n          SpawnSdkExample.Actors.JoeActor\n        ],\n        extenal_subscribers: [\n          {SpawnSdkExample.Subscriber, []}\n        ]\n      }\n    ]\n\n    opts = [strategy: :one_for_one, name: SpawnSdkExample.Supervisor]\n    Supervisor.start_link(children, opts)\n  end\nend\n```\n\nThe important thing here is to use the external_subscribers attribute. As seen above :external_subscribers accepts a list of specs as a parameter.\n\n3. Set your actor as you normally would and emit your broadcast events using Broadcast.to(channel, payload).\n\n```elixir\ndefmodule SpawnSdkExample.Actors.JoeActor do\n  use SpawnSdk.Actor,\n    name: \"joe\",\n    state_type: Io.Eigr.Spawn.Example.MyState\n\n  require Logger\n  alias Io.Eigr.Spawn.Example.{MyState, MyBusinessMessage}\n\n  defact sum(%MyBusinessMessage{value: value} = data, %Context{state: state} = ctx) do\n    Logger.info(\"[joe] Received Request: #{inspect(data)}. Context: #{inspect(ctx)}\")\n\n    new_value =\n      if is_nil(state) do\n        0 + value\n      else\n        (state.value || 0) + value\n      end\n\n    response = %MyBusinessMessage{value: new_value}\n\n    %Value{}\n    |> Value.of(response, %MyState{value: new_value})\n    |> Value.broadcast(Broadcast.to(\"my.channel\", response))\n  end\nend\n```","ref":"readme.html#broadcast-to-external-subscribers","title":"Broadcast to External Subscribers - Spawn Elixir SDK","type":"extras"},{"doc":"Actors can also declare Actions that act recursively as timers. See an example below:\n\n```elixir\ndefmodule SpawnSdkExample.Actors.ClockActor do\n  use SpawnSdk.Actor,\n    name: \"clock_actor\",\n    state_type: Io.Eigr.Spawn.Example.MyState,\n    deactivate_timeout: 86_400_000\n\n  require Logger\n\n  alias Io.Eigr.Spawn.Example.MyState\n\n  @set_timer 15_000\n  defact clock(%Context{state: state} = ctx) do\n    Logger.info(\"[clock] Clock Actor Received Request. Context: #{inspect(ctx)}\")\n\n    new_value = if is_nil(state), do: 0, else: state.value + 1\n    new_state = MyState.new(value: new_value)\n\n    Value.of()\n    |> Value.state(new_state)\n  end\nend\n```\n\n> **_NOTE:_** Timers Actions are recorded as Actor metadata. Where in turn we use a synchronization mechanism via CRDTs to keep the metadata alive in the cluster while there is an active Spawn VM. That is, Timers Actions are ephemeral and therefore only exist while there is at least one active VM in the cluster.\n\nIn the example above the ´clock´ action will be called every 15 seconds.","ref":"readme.html#timers","title":"Timers - Spawn Elixir SDK","type":"extras"},{"doc":"Once we define our actors we can now declare our supervision tree:\n\n```elixir\ndefmodule SpawnSdkExample.Application do\n  @moduledoc false\n  use Application\n\n  @impl true\n  def start(_type, _args) do\n    children = [\n      {\n        SpawnSdk.System.Supervisor,\n        system: \"spawn-system\",\n        actors: [\n          SpawnSdkExample.Actors.MyActor,\n          SpawnSdkExample.Actors.UnamedActor,\n          SpawnSdkExample.Actors.ClockActor,\n          SpawnSdkExample.Actors.PooledActor\n        ]\n      }\n    ]\n\n    opts = [strategy: :one_for_one, name: SpawnSdkExample.Supervisor]\n    Supervisor.start_link(children, opts)\n  end\nend\n```","ref":"readme.html#declaring-the-supervision-tree","title":"Declaring the supervision tree - Spawn Elixir SDK","type":"extras"},{"doc":"Actors also have some standard actions that are not implemented by the user and that can be used as a way to get the state of an actor without the invocation requiring an extra trip to the host functions. You can think of them as a cache of their state, every time you invoke a default action on an actor it will return the value directly from the Sidecar process without this target process needing to invoke its equivalent host function.\n\nLet's take an example. Suppose Actor Joe wants to know the current state of Actor Robert. What Joe can do is invoke Actor Robert's default action called get_state. This will make Actor Joe's sidecar find Actor Robert's sidecar somewhere in the cluster and Actor Robert's sidecar will return its own state directly to Joe without having to resort to your host function, this in turn will save you a called over the network and therefore this type of invocation is faster than invocations of user-defined actions usually are.\n\nAny invocations to actions with the following names will follow this rule: \"**get**\",\n\"**Get**\",\n\"**get_state**\",\n\"**getState**\",\n\"**GetState**\"\n\n> **_NOTE:_** You can override this behavior by defining your actor as an action with the same name as the default actions. In this case it will be the Action defined by you that will be called, implying perhaps another network roundtrip","ref":"readme.html#default-actions","title":"Default Actions - Spawn Elixir SDK","type":"extras"},{"doc":"To deploy your actors in a cluster you need to use our Controller for Kubernetes. Detailed information on how to proceed can be found [here](https://github.com/eigr/spawn#install) and [here](https://github.com/eigr/spawn#getting-started).\nBut you can also run your Elixir application in the traditional way as follows:\n\n```\nMIX_ENV=prod USER_FUNCTION_PORT=8092 PROXY_DATABASE_TYPE=mysql SPAWN_STATESTORE_KEY=3Jnb0hZiHIzHTOih7t2cTEPEpY98Tu1wvQkPfq/XwqE= iex --name spawn_a2@127.0.0.1 -S mix\n```\n\n> **_NOTE:_** For reasons of compatibility with our controller, it is necessary to configure your Spawn application using environment variables instead of the traditional Elixir configuration mechanism.\n\n> **_WARNING:_** This example uses the MySQL database as persistent storage for its actors. And it is also expected that you have previously created a database called eigr-functions-db in the MySQL instance.\n\nThe full example of this application can be found [here](https://github.com/eigr/spawn/tree/main/spawn_sdk/spawn_sdk_example).\n\nAnd links to other examples can be found in our github [readme page](https://github.com/eigr/spawn#examples).","ref":"readme.html#running","title":"Running - Spawn Elixir SDK","type":"extras"},{"doc":"To invoke Actors, use:\n\n```elixir\niex> SpawnSdk.invoke(\"joe\", system: \"spawn-system\", action: \"sum\", payload: %Io.Eigr.Spawn.Example.MyBusinessMessage{value: 1})\n{:ok, %Io.Eigr.Spawn.Example.MyBusinessMessage{value: 12}}\n```\n\nYou can invoke actor default functions like \"get\" to get its current state\n\n```elixir\nSpawnSdk.invoke(\"joe\", system: \"spawn-system\", action: \"get\")\n```\n\nSpawning Actors:\n\n```elixir\niex> SpawnSdk.spawn_actor(\"robert\", system: \"spawn-system\", actor: \"unamed_actor\")\n:ok\n```\n\nYou can also create Actors so that they are initialized from a certain revision number, that is, initialize actors from a specific point in time.\n\n```elixir\niex> SpawnSdk.spawn_actor(\"robert\", system: \"spawn-system\", actor: \"unamed_actor\", revision: 2)\n:ok\n```\n\nIn the above case the actor will be initialized with its state restored from the state as it was in revision 2 of its previous lifetime.\n\nInvoke Spawned Actors:\n\n```elixir\niex> SpawnSdk.invoke(\"robert\", system: \"spawn-system\", action: \"sum\", payload: %Io.Eigr.Spawn.Example.MyBusinessMessage{value: 1})\n{:ok, %Io.Eigr.Spawn.Example.MyBusinessMessage{value: 16}}\n```\n\nInvoke Actors in a lazy way without having to spawn them before:\n\n```elixir\niex> SpawnSdk.invoke(\"robert_lazy\", ref: SpawnSdkExample.Actors.UnamedActor, system: \"spawn-system\", action: \"sum\", payload: %Io.Eigr.Spawn.Example.MyBusinessMessage{value: 1})\n{:ok, %Io.Eigr.Spawn.Example.MyBusinessMessage{value: 1}}\n```\n\nInvoke Actors with a delay set in milliseconds:\n\n```elixir\niex> SpawnSdk.invoke(\"joe\", system: \"spawn-system\", action: \"ping\", delay: 5_000)\n{:ok, :async}\n```\n\nInvoke Actors scheduled to a specific DateTime:\n\n```elixir\niex> SpawnSdk.invoke(\"joe\", system: \"spawn-system\", action: \"ping\", scheduled_to: ~U[2023-01-01 00:32:00.145Z])\n{:ok, :async}\n```\n\nInvoke Pooled Actors:\n\n```elixir\niex> SpawnSdk.invoke(\"pooled_actor\", system: \"spawn-system\", action: \"ping\", pooled: true)\n{:ok, nil}\n```","ref":"readme.html#client-api-examples","title":"Client API Examples - Spawn Elixir SDK","type":"extras"}]}