searchNodes=[{"doc":"Spawn Elixir SDK is the support library for the Spawn Actors System. Spawn is a Stateful Serverless Platform for providing the multi-language Actor Model. For a broader understanding of Spawn, please consult its official repository . The advantage of the Elixir SDK over other SDKs is in Elixir's native ability to connect directly to an Erlang network. For this reason, the Elixir SDK allows any valid Elixir application to be part of a Spawn network without needing a sidecar attached. Installation Available in Hex , the package can be installed by adding spawn_sdk and spawn_statestores_* to your list of dependencies in mix.exs : def deps do [ { :spawn_sdk , &quot;~&gt; 1.0.0-rc1&quot; } , # You can uncomment one of those dependencies if you are going to use Persistent Actors \#{:spawn_statestores_mysql, &quot;~&gt; 1.0.0-rc1&quot;}, \#{:spawn_statestores_postgres, &quot;~&gt; 1.0.0-rc1&quot;}, \#{:spawn_statestores_mssql, &quot;~&gt; 1.0.0-rc1&quot;}, \#{:spawn_statestores_cockroachdb, &quot;~&gt; 1.0.0-rc1&quot;}, \#{:spawn_statestores_sqlite, &quot;~&gt; 1.0.0-rc1&quot;}, ] end Deploy Following the steps below, you will have a valid Elixir application to use in a Spawn cluster. However, you will still need to generate a container image with your application to use it together with the Spawn Operator for Kubernetes . How to use After creating an Elixir application project, create the protobuf files for your business domain. It is common practice to do this under the priv/ folder. Let's demonstrate an example: syntax = &quot;proto3&quot;; package io.eigr.spawn.example; message MyState { int32 value = 1; } message MyBusinessMessage { int32 value = 1; } It is important to try to separate the type of message that must be stored as the actors' state from the type of messages that will be exchanged between their actors' operations calls. In other words, the Actor's internal state is also represented as a protobuf type, and it is a good practice to separate these types of messages from the others in its business domain. In the above case MyState is the type protobuf that represents the state of the Actor that we will create later while MyBusiness is the type of message that we will send and receive from this Actor. Now that we have defined our input and output types as Protobuf types we will need to compile these files to generate their respective Elixir modules. An example of how to do this can be found here NOTE: You need to have installed the elixir plugin for protoc. More information on how to obtain and install the necessary tools can be found here here Now that the protobuf types have been created we can proceed with the code. Example definition of an Actor. Singleton Actors In this example we are creating an actor in a Singleton way, that is, it is a known actor at compile time. defmodule SpawnSdkExample.Actors.MyActor do use SpawnSdk.Actor , name : &quot;jose&quot; , # Default is Full Qualified Module name a.k.a __MODULE__ kind : :singleton , # Default is already :singleton. Valid are :singleton | :abstract | :pooled stateful : true , # Default is already true state_type : Io.Eigr.Spawn.Example.MyState , # or :json if you don&#39;t care about protobuf types deactivate_timeout : 30_000 , snapshot_timeout : 2_000 require Logger alias Io.Eigr.Spawn.Example . { MyState , MyBusinessMessage } defact init ( % Context { state : state } = ctx ) do Logger . info ( &quot;[joe] Received InitRequest. Context: \#{ inspect ( ctx ) } &quot; ) Value . of ( ) |&gt; Value . state ( state ) end defact sum ( % MyBusinessMessage { value : value } = data , % Context { state : state } = ctx ) do Logger . info ( &quot;Received Request: \#{ inspect ( data ) } . Context: \#{ inspect ( ctx ) } &quot; ) new_value = if is_nil ( state ) , do : value , else : ( state . value || 0 ) + value Value . of ( % MyBusinessMessage { value : new_value } , % MyState { value : new_value } ) end end We declare two actions that the Actor can do. An initialization action that will be called every time an Actor instance is created and an action that will be responsible for performing a simple sum. Note Keep in mind that any Action that has the names present in the list below will behave as an initialization Action and will be called when the Actor is started (if there is more than one Action with one of these names, only one will be called). Defaults inicialization Action names: &quot; init &quot;, &quot; Init &quot;, &quot; setup &quot;, &quot; Setup &quot; Abstract Actor We can also create Unnamed Dynamic/Lazy actors, that is, despite having its abstract behavior defined at compile time, a Lazy actor will only have a concrete instance when it is associated with an identifier/name at runtime. Below follows the same previous actor being defined as abstract. defmodule SpawnSdkExample.Actors.AbstractActor do use SpawnSdk.Actor , name : &quot;abs_actor&quot; , kind : :abstract , state_type : Io.Eigr.Spawn.Example.MyState require Logger alias Io.Eigr.Spawn.Example . { MyState , MyBusinessMessage } defact sum ( % MyBusinessMessage { value : value } = data , % Context { state : state } = ctx ) do Logger . info ( &quot;Received Request: \#{ inspect ( data ) } . Context: \#{ inspect ( ctx ) } &quot; ) new_value = if is_nil ( state ) , do : value , else : ( state . value || 0 ) + value Value . of ( % MyBusinessMessage { value : new_value } , % MyState { value : new_value } ) end end Notice that the only thing that has changed is the the kind of actor, in this case the kind is set to :abstract. NOTE: Can Elixir programmers think in terms of singleton vs abstract actors as more or less known at startup vs dynamically supervised/registered? That is, defining your actors directly in the supervision tree or using a Dynamic Supervisor for that. Pooled Actors Sometimes we want a particular actor to be able to serve requests concurrently, however actors will always serve one request at a time using buffering mechanisms to receive requests in their mailbox and serve each request one by one. So to get around this behaviour you can configure your Actor as a Pooled Actor, this way the system will generate a pool of actors to meet certain requests. See an example below: defmodule SpawnSdkExample.Actors.PooledActor do use SpawnSdk.Actor , name : &quot;pooled_actor&quot; , kind : :pooled , stateful : false require Logger defact ping ( _data , % Context { } = ctx ) do Logger . info ( &quot;Received Request. Context: \#{ inspect ( ctx ) } &quot; ) Value . void ( ) end end NOTE: It is important to mention that Pooled Actors cannot be stateful because concurrent accesses for Stateful actors could generate undesired inconsistencies. To invoke a pooled actor, it is necessary to pass the pooled: true option to the invoke function. See an example: iex ( spawn_a1 @ 127.0 . 0.1 ) 16 &gt; for _n &lt;- 0 .. 8 , do : SpawnSdk . invoke ( &quot;pooled_actor&quot; , system : &quot;spawn-system&quot; , command : &quot;ping&quot; , pooled : true ) In the logs you will see that even though you have invoked the actor by the registered name more than one instance of the actor will be created. As we can see below: iex ( spawn_a1 @ 127.0 . 0.1 ) 16 &gt; for _n &lt;- 0 .. 8 , do : SpawnSdk . invoke ( &quot;pooled_actor&quot; , system : &quot;spawn-system&quot; , command : &quot;ping&quot; , pooled : true ) 2022 - 11 - 25 11 : 21 : 53.758 [ spawn_a1 @ 127.0 . 0.1 ] : [ pid = &lt; 0.1392 . 0 &gt; ] : [ notice ] :Activating Actor pooled_actor - 5 with Parent pooled_actor - 5 in Node :&quot;spawn_a1@127.0.0.1&quot; . Persistence false . 2022 - 11 - 25 11 : 21 : 53.759 [ spawn_a1 @ 127.0 . 0.1 ] : [ pid = &lt; 0.1391 . 0 &gt; ] : [ debug ] :Actor pooled_actor - 5 reactivated . ActorRef PID : # PID &lt; 0 . 1392 . 0 &gt; 2022 - 11 - 25 11 : 21 : 53.760 [ spawn_a1 @ 127.0 . 0.1 ] : [ pid = &lt; 0.1392 . 0 &gt; ] : [ info ] :Received Request . Context : % SpawnSdk.Context { state : nil , caller : nil , self : % Eigr.Functions.Protocol.Actors.ActorId { name : &quot;pooled_actor-5&quot; , system : &quot;spawn-system&quot; , parent : &quot;pooled_actor-5&quot; , __unknown_fields__ : [ ] } , metadata : %{ } } 2022 - 11 - 25 11 : 21 : 53.763 [ spawn_a1 @ 127.0 . 0.1 ] : [ pid = &lt; 0.1394 . 0 &gt; ] : [ notice ] :Activating Actor pooled_actor - 4 with Parent pooled_actor - 4 in Node :&quot;spawn_a1@127.0.0.1&quot; . Persistence false . 2022 - 11 - 25 11 : 21 : 53.763 [ spawn_a1 @ 127.0 . 0.1 ] : [ pid = &lt; 0.1393 . 0 &gt; ] : [ debug ] :Actor pooled_actor - 4 reactivated . ActorRef PID : # PID &lt; 0 . 1394 . 0 &gt; 2022 - 11 - 25 11 : 21 : 53.765 [ spawn_a1 @ 127.0 . 0.1 ] : [ pid = &lt; 0.1394 . 0 &gt; ] : [ info ] :Received Request . Context : % SpawnSdk.Context { state : nil , caller : nil , self : % Eigr.Functions.Protocol.Actors.ActorId { name : &quot;pooled_actor-4&quot; , system : &quot;spawn-system&quot; , parent : &quot;pooled_actor-4&quot; , __unknown_fields__ : [ ] } , metadata : %{ } } 2022 - 11 - 25 11 : 21 : 53.767 [ spawn_a1 @ 127.0 . 0.1 ] : [ pid = &lt; 0.1396 . 0 &gt; ] : [ notice ] :Activating Actor pooled_actor - 7 with Parent pooled_actor - 7 in Node :&quot;spawn_a1@127.0.0.1&quot; . Persistence false . 2022 - 11 - 25 11 : 21 : 53.767 [ spawn_a1 @ 127.0 . 0.1 ] : [ pid = &lt; 0.1395 . 0 &gt; ] : [ debug ] :Actor pooled_actor - 7 reactivated . ActorRef PID : # PID &lt; 0 . 1396 . 0 &gt; 2022 - 11 - 25 11 : 21 : 53.768 [ spawn_a1 @ 127.0 . 0.1 ] : [ pid = &lt; 0.1396 . 0 &gt; ] : [ info ] :Received Request . Context : % SpawnSdk.Context { state : nil , caller : nil , self : % Eigr.Functions.Protocol.Actors.ActorId { name : &quot;pooled_actor-7&quot; , system : &quot;spawn-system&quot; , parent : &quot;pooled_actor-7&quot; , __unknown_fields__ : [ ] } , metadata : %{ } } 2022 - 11 - 25 11 : 21 : 53.770 [ spawn_a1 @ 127.0 . 0.1 ] : [ pid = &lt; 0.1398 . 0 &gt; ] : [ notice ] :Activating Actor pooled_actor - 2 with Parent pooled_actor - 2 in Node :&quot;spawn_a1@127.0.0.1&quot; . Persistence false . 2022 - 11 - 25 11 : 21 : 53.771 [ spawn_a1 @ 127.0 . 0.1 ] : [ pid = &lt; 0.1397 . 0 &gt; ] : [ debug ] :Actor pooled_actor - 2 reactivated . ActorRef PID : # PID &lt; 0 . 1398 . 0 &gt; 2022 - 11 - 25 11 : 21 : 53.772 [ spawn_a1 @ 127.0 . 0.1 ] : [ pid = &lt; 0.1398 . 0 &gt; ] : [ info ] :Received Request . Context : % SpawnSdk.Context { state : nil , caller : nil , self : % Eigr.Functions.Protocol.Actors.ActorId { name : &quot;pooled_actor-2&quot; , system : &quot;spawn-system&quot; , parent : &quot;pooled_actor-2&quot; , __unknown_fields__ : [ ] } , metadata : %{ } } 2022 - 11 - 25 11 : 21 : 53.774 [ spawn_a1 @ 127.0 . 0.1 ] : [ pid = &lt; 0.1400 . 0 &gt; ] : [ notice ] :Activating Actor pooled_actor - 8 with Parent pooled_actor - 8 in Node :&quot;spawn_a1@127.0.0.1&quot; . Persistence false . 2022 - 11 - 25 11 : 21 : 53.775 [ spawn_a1 @ 127.0 . 0.1 ] : [ pid = &lt; 0.1399 . 0 &gt; ] : [ debug ] :Actor pooled_actor - 8 reactivated . ActorRef PID : # PID &lt; 0 . 1400 . 0 &gt; 2022 - 11 - 25 11 : 21 : 53.776 [ spawn_a1 @ 127.0 . 0.1 ] : [ pid = &lt; 0.1400 . 0 &gt; ] : [ info ] :Received Request . Context : % SpawnSdk.Context { state : nil , caller : nil , self : % Eigr.Functions.Protocol.Actors.ActorId { name : &quot;pooled_actor-8&quot; , system : &quot;spawn-system&quot; , parent : &quot;pooled_actor-8&quot; , __unknown_fields__ : [ ] } , metadata : %{ } } 2022 - 11 - 25 11 : 21 : 53.777 [ spawn_a1 @ 127.0 . 0.1 ] : [ pid = &lt; 0.1402 . 0 &gt; ] : [ notice ] :Activating Actor pooled_actor - 9 with Parent pooled_actor - 9 in Node :&quot;spawn_a1@127.0.0.1&quot; . Persistence false . 2022 - 11 - 25 11 : 21 : 53.778 [ spawn_a1 @ 127.0 . 0.1 ] : [ pid = &lt; 0.1401 . 0 &gt; ] : [ debug ] :Actor pooled_actor - 9 reactivated . ActorRef PID : # PID &lt; 0 . 1402 . 0 &gt; 2022 - 11 - 25 11 : 21 : 53.778 [ spawn_a1 @ 127.0 . 0.1 ] : [ pid = &lt; 0.1402 . 0 &gt; ] : [ info ] :Received Request . Context : % SpawnSdk.Context { state : nil , caller : nil , self : % Eigr.Functions.Protocol.Actors.ActorId { name : &quot;pooled_actor-9&quot; , system : &quot;spawn-system&quot; , parent : &quot;pooled_actor-9&quot; , __unknown_fields__ : [ ] } , metadata : %{ } } 2022 - 11 - 25 11 : 21 : 53.779 [ spawn_a1 @ 127.0 . 0.1 ] : [ pid = &lt; 0.1212 . 0 &gt; ] : [ debug ] :Lookup Actor pooled_actor . PID : # PID &lt; 0 . 1400 . 0 &gt; 2022 - 11 - 25 11 : 21 : 53.780 [ spawn_a1 @ 127.0 . 0.1 ] : [ pid = &lt; 0.1400 . 0 &gt; ] : [ info ] :Received Request . Context : % SpawnSdk.Context { state : nil , caller : nil , self : % Eigr.Functions.Protocol.Actors.ActorId { name : &quot;pooled_actor-8&quot; , system : &quot;spawn-system&quot; , parent : &quot;pooled_actor-8&quot; , __unknown_fields__ : [ ] } , metadata : %{ } } 2022 - 11 - 25 11 : 21 : 53.780 [ spawn_a1 @ 127.0 . 0.1 ] : [ pid = &lt; 0.1212 . 0 &gt; ] : [ debug ] :Lookup Actor pooled_actor . PID : # PID &lt; 0 . 1392 . 0 &gt; 2022 - 11 - 25 11 : 21 : 53.781 [ spawn_a1 @ 127.0 . 0.1 ] : [ pid = &lt; 0.1392 . 0 &gt; ] : [ info ] :Received Request . Context : % SpawnSdk.Context { state : nil , caller : nil , self : % Eigr.Functions.Protocol.Actors.ActorId { name : &quot;pooled_actor-5&quot; , system : &quot;spawn-system&quot; , parent : &quot;pooled_actor-5&quot; , __unknown_fields__ : [ ] } , metadata : %{ } } 2022 - 11 - 25 11 : 21 : 53.781 [ spawn_a1 @ 127.0 . 0.1 ] : [ pid = &lt; 0.1212 . 0 &gt; ] : [ debug ] :Lookup Actor pooled_actor . PID : # PID &lt; 0 . 1394 . 0 &gt; [ ok : nil , ok : nil , ok : nil , ok : nil , ok : nil , ok : nil , ok : nil , ok : nil , ok : nil ] iex ( spawn_a1 @ 127.0 . 0.1 ) 17 &gt; 2022 - 11 - 25 11 : 21 : 53.782 [ spawn_a1 @ 127.0 . 0.1 ] : [ pid = &lt; 0.1394 . 0 &gt; ] : [ info ] :Received Request . Context : % SpawnSdk.Context { state : nil , caller : nil , self : % Eigr.Functions.Protocol.Actors.ActorId { name : &quot;pooled_actor-4&quot; , system : &quot;spawn-system&quot; , parent : &quot;pooled_actor-4&quot; , __unknown_fields__ : [ ] } , metadata : %{ } } 2022 - 11 - 25 11 : 22 : 03.999 [ spawn_a1 @ 127.0 . 0.1 ] : [ pid = &lt; 0.1402 . 0 &gt; ] : [ debug ] :Deactivating actor pooled_actor - 9 for timeout 2022 - 11 - 25 11 : 22 : 04.226 [ spawn_a1 @ 127.0 . 0.1 ] : [ pid = &lt; 0.1392 . 0 &gt; ] : [ debug ] :Deactivating actor pooled_actor - 5 for timeout 2022 - 11 - 25 11 : 22 : 07.271 [ spawn_a1 @ 127.0 . 0.1 ] : [ pid = &lt; 0.1400 . 0 &gt; ] : [ debug ] :Deactivating actor pooled_actor - 8 for timeout 2022 - 11 - 25 11 : 22 : 10.167 [ spawn_a1 @ 127.0 . 0.1 ] : [ pid = &lt; 0.1396 . 0 &gt; ] : [ debug ] :Deactivating actor pooled_actor - 7 for timeout 2022 - 11 - 25 11 : 22 : 12.038 [ spawn_a1 @ 127.0 . 0.1 ] : [ pid = &lt; 0.1394 . 0 &gt; ] : [ debug ] :Deactivating actor pooled_actor - 4 for timeout 2022 - 11 - 25 11 : 22 : 12.613 [ spawn_a1 @ 127.0 . 0.1 ] : [ pid = &lt; 0.1398 . 0 &gt; ] : [ debug ] :Deactivating actor pooled_actor - 2 for timeout It is possible to configure the minimum and maximum pool size by passing the min_pool_size and max_pool_size options. The actor below would create a pool of a maximum of 100 Actors: defmodule SpawnSdkExample.Actors.PooledActor do use SpawnSdk.Actor , name : &quot;pooled_actor&quot; , kind : :pooled , min_pool_size : 1 , max_pool_size : 10 , stateful : false # Code omitted for brevity end ALERT: Keep in mind that a large pool size does not always mean better performance. In fact, smaller pool sizes sometimes often have a much greater effect in this regard. It is interesting that you experiment with different pool sizes until you find the one that best suits your workload. Side Effects Actors can also emit side effects to other Actors as part of their response. See an example: defmodule SpawnSdkExample.Actors.AbstractActor do use SpawnSdk.Actor , kind : :abstract , stateful : false , state_type : Io.Eigr.Spawn.Example.MyState require Logger alias Io.Eigr.Spawn.Example . { MyState , MyBusinessMessage } alias SpawnSdk.Flow.SideEffect defact sum ( % MyBusinessMessage { value : value } = data , % Context { state : state } = ctx ) do Logger . info ( &quot;Received Request: \#{ inspect ( data ) } . Context: \#{ inspect ( ctx ) } &quot; ) new_value = if is_nil ( state ) , do : value , else : ( state . value || 0 ) + value result = % MyBusinessMessage { value : new_value } new_state = % MyState { value : new_value } Value . of ( ) |&gt; Value . response ( result ) |&gt; Value . state ( new_state ) |&gt; Value . effects ( # This returns a list of side effects. In this case containing only one effect. However, multiple effects can be chained together, # just by calling the effect function as shown here. # The delay means that it will be fired asynchronously after 5000 milliseconds (5 seconds) # If only one effect is desired, you can also choose to use the to/3 function together with Value.effect(). # Example: Values.effect(SideEffect.to(name, func, payload)) SideEffect . of ( ) |&gt; SideEffect . effect ( &quot;joe&quot; , :sum , result , delay : 5_000 , scheduled_to : ~U[2020-01-01 10:00:00.145Z] ) # use delay or scheduled_to, not both ) end end In the example above we see that the Actor joe will receive a request as a side effect from the Actor who issued this effect. Side effects do not interfere with an actor's request-response flow. They will &quot;always&quot; be processed asynchronously and any response sent back from the Actor receiving the effect will be ignored by the effector. Pipe and Forward Actors can also route some commands to other actors as part of their response. See an example: defmodule SpawnSdkExample.Actors.ForwardPipeActor do use SpawnSdk.Actor , name : &quot;pipeforward&quot; , kind : :singleton , stateful : false require Logger alias Io.Eigr.Spawn.Example.MyBusinessMessage defact forward_example ( % MyBusinessMessage { } = msg , _ctx ) do Logger . info ( &quot;Received request with \#{ msg . value } &quot; ) Value . of ( ) |&gt; Value . response ( MyBusinessMessage . new ( value : 999 ) ) |&gt; Value . forward ( Forward . to ( &quot;second_actor&quot; , &quot;sum_plus_one&quot; ) ) |&gt; Value . void ( ) end defact pipe_example ( % MyBusinessMessage { } = msg , _ctx ) do Logger . info ( &quot;Received request with \#{ msg . value } &quot; ) Value . of ( ) |&gt; Value . response ( MyBusinessMessage . new ( value : 999 ) ) |&gt; Value . pipe ( Pipe . to ( &quot;second_actor&quot; , &quot;sum_plus_one&quot; ) ) |&gt; Value . void ( ) end end defmodule SpawnSdkExample.Actors.SecondActorExample do use SpawnSdk.Actor , name : &quot;second_actor&quot; , stateful : false require Logger alias Io.Eigr.Spawn.Example.MyBusinessMessage defact sum_plus_one ( % MyBusinessMessage { } = msg , _ctx ) do Logger . info ( &quot;Received request with \#{ msg . value } &quot; ) Value . of ( ) |&gt; Value . response ( MyBusinessMessage . new ( value : msg . value + 1 ) ) |&gt; Value . void ( ) end end We are returning void in both examples so we dont care about what is being stored in the actor state. In the case above, every time you call the forward_example the second_actor's sum_plus_one function will receive the value forwarded originally in the invocation as its input. The end result will be: iex&gt; SpawnSdk . invoke ( &quot;pipeforward&quot; , system : &quot;spawn-system&quot; , command : &quot;forward_example&quot; , payload : % Io.Eigr.Spawn.Example.MyBusinessMessage { value : 1 } ) { :ok , % Io.Eigr.Spawn.Example.MyBusinessMessage { value : 2 } } For the Pipe example, the the second_actor's sum_plus_one function will always receive %MyBusinessMessage{value: 999} due to getting the value from the previous specification in the pipe_example command, the end result will be: iex&gt; SpawnSdk . invoke ( &quot;pipeforward&quot; , system : &quot;spawn-system&quot; , command : &quot;pipe_example&quot; , payload : % Io.Eigr.Spawn.Example.MyBusinessMessage { value : 1 } ) { :ok , % Io.Eigr.Spawn.Example.MyBusinessMessage { value : 1000 } } Broadcast Actors can also send messages to a group of actors at once as an action callback. See the example below: defmodule Fleet.Actors.Driver do use SpawnSdk.Actor , kind : :abstract , # Set ´driver´ channel for all actors of the same type (Fleet.Actors.Driver) channel : &quot;drivers&quot; , state_type : Fleet.Domain.Driver alias Fleet.Domain . { Driver , OfferRequest , OfferResponse , Point } require Logger @brain_actor_channel &quot;fleet-controllers&quot; defact update_position ( % Point { } = position , % Context { state : % Driver { id : name } = driver } = ctx ) do Logger . info ( &quot;Driver [ \#{ name } ] Received Update Position Event. Position: [ \#{ inspect ( position ) } ]. Context: \#{ inspect ( ctx ) } &quot; ) driver_state = % Driver { driver | position : position } % Value { } |&gt; Value . of ( driver_state , driver_state ) |&gt; Value . broadcast ( Broadcast . to ( @brain_actor_channel , &quot;driver_position&quot; , driver_state ) ) end end In the case above, every time an Actor &quot;driver&quot; executes the update_position action it will send a message to all the actors participating in the channel called &quot;fleet-controllers&quot;. Broadcast to External Subscribers Sometimes you may want to send events out of ActorSystem using Phoenix.PubSub. One way to do this is to take advantage of the same Broadcast infrastructure that Spawn offers you but indicating an external channel. Below is an example: Create a Listener to receive the events using the SpawnSdk.Channel.Subscriber helper module. defmodule SpawnSdkExample.Subscriber do @moduledoc &quot;&quot;&quot; This module exemplifies how to listen for pubsub events that were emitted by actors but that will be treated not by actors but as normal pubsub events. This is particularly useful for integrations between Spawn and Phoenix LiveView. &quot;&quot;&quot; use GenServer require logger alias SpawnSdk.Channel.Subscriber @impl true define init ( state ) do Subscriber . subscribe ( &quot;external.channel&quot; ) { :ok , state } end @impl true def handle_info ( { :receive , payload } , state ) do Logger . info ( &quot;Received pubsub event \#{ inspect ( payload ) } &quot; ) { :noreply , state } end def start_link ( args ) do GenServer . start_link ( __MODULE__ , args ) end end You need to match using the {:receive, payload} tuple in your handle_info. NOTE: By default SpawnSdk.Channel.Subscriber will subscribe to pubsub using the atom :actor_channel as an argument. If you need to change this , just configure your configuration as follows : config.exs config :spawn , pubsub_group : :your_channel_group_here SpawnSdk.System.Supervisor. defmodule SpawnSdkExample.Application do @moduledoc false use Application @impl true def start ( _type , _args ) do children = [ { SpawnSdk.System.Supervisor , system : &quot;spawn-system&quot; , actors : [ SpawnSdkExample.Actors.JoeActor ] , extenal_subscribers : [ { SpawnSdkExample.Subscriber , [ ] } ] } ] opts = [ strategy : :one_for_one , name : SpawnSdkExample.Supervisor ] Supervisor . start_link ( children , opts ) end end The important thing here is to use the external_subscribers attribute. As seen above :external_subscribers accepts a list of specs as a parameter. Set your actor as you normally would and emit your broadcast events using Broadcast.to(channel, payload). defmodule SpawnSdkExample.Actors.JoeActor do use SpawnSdk.Actor , name : &quot;joe&quot; , state_type : Io.Eigr.Spawn.Example.MyState require Logger alias Io.Eigr.Spawn.Example . { MyState , MyBusinessMessage } defact sum ( % MyBusinessMessage { value : value } = data , % Context { state : state } = ctx ) do Logger . info ( &quot;[joe] Received Request: \#{ inspect ( data ) } . Context: \#{ inspect ( ctx ) } &quot; ) new_value = if is_nil ( state ) do 0 + value else ( state . value || 0 ) + value end response = % MyBusinessMessage { value : new_value } % Value { } |&gt; Value . of ( response , % MyState { value : new_value } ) |&gt; Value . broadcast ( Broadcast . to ( &quot;my.channel&quot; , response ) ) end end Timers Actors can also declare Actions that act recursively as timers. See an example below: defmodule SpawnSdkExample.Actors.ClockActor do use SpawnSdk.Actor , name : &quot;clock_actor&quot; , state_type : Io.Eigr.Spawn.Example.MyState , deactivate_timeout : 86_400_000 require Logger alias Io.Eigr.Spawn.Example.MyState @set_timer 15_000 defact clock ( % Context { state : state } = ctx ) do Logger . info ( &quot;[clock] Clock Actor Received Request. Context: \#{ inspect ( ctx ) } &quot; ) new_value = if is_nil ( state ) , do : 0 , else : state . value + 1 new_state = MyState . new ( value : new_value ) Value . of ( ) |&gt; Value . state ( new_state ) end end NOTE: Timers Actions are recorded as Actor metadata. Where in turn we use a synchronization mechanism via CRDTs to keep the metadata alive in the cluster while there is an active Spawn VM. That is, Timers Actions are ephemeral and therefore only exist while there is at least one active VM in the cluster. In the example above the ´clock´ action will be called every 15 seconds. Declaring the supervision tree Once we define our actors we can now declare our supervision tree: defmodule SpawnSdkExample.Application do @moduledoc false use Application @impl true def start ( _type , _args ) do children = [ { SpawnSdk.System.Supervisor , system : &quot;spawn-system&quot; , actors : [ SpawnSdkExample.Actors.MyActor , SpawnSdkExample.Actors.AbstractActor , SpawnSdkExample.Actors.ClockActor , SpawnSdkExample.Actors.PooledActor ] } ] opts = [ strategy : :one_for_one , name : SpawnSdkExample.Supervisor ] Supervisor . start_link ( children , opts ) end end Default Actions Actors also have some standard actions that are not implemented by the user and that can be used as a way to get the state of an actor without the invocation requiring an extra trip to the host functions. You can think of them as a cache of their state, every time you invoke a default action on an actor it will return the value directly from the Sidecar process without this target process needing to invoke its equivalent host function. Let's take an example. Suppose Actor Joe wants to know the current state of Actor Robert. What Joe can do is invoke Actor Robert's default action called get_state. This will make Actor Joe's sidecar find Actor Robert's sidecar somewhere in the cluster and Actor Robert's sidecar will return its own state directly to Joe without having to resort to your host function, this in turn will save you a called over the network and therefore this type of invocation is faster than invocations of user-defined actions usually are. Any invocations to actions with the following names will follow this rule: &quot; get &quot;, &quot; Get &quot;, &quot; get_state &quot;, &quot; getState &quot;, &quot; GetState &quot; NOTE: You can override this behavior by defining your actor as an action with the same name as the default actions. In this case it will be the Action defined by you that will be called, implying perhaps another network roundtrip Running To deploy your actors in a cluster you need to use our Controller for Kubernetes. Detailed information on how to proceed can be found here and here . But you can also run your Elixir application in the traditional way as follows: MIX_ENV = prod USER_FUNCTION_PORT = 8092 PROXY_DATABASE_TYPE = mysql SPAWN_STATESTORE_KEY = 3 Jnb0hZiHIzHTOih7t2cTEPEpY98Tu1wvQkPfq / XwqE = iex -- name spawn_a2 @ 127.0 . 0.1 - S mix NOTE: For reasons of compatibility with our controller, it is necessary to configure your Spawn application using environment variables instead of the traditional Elixir configuration mechanism. WARNING: This example uses the MySQL database as persistent storage for its actors. And it is also expected that you have previously created a database called eigr-functions-db in the MySQL instance. The full example of this application can be found here . And links to other examples can be found in our github readme page . Client API Examples To invoke Actors, use: iex&gt; SpawnSdk . invoke ( &quot;joe&quot; , system : &quot;spawn-system&quot; , command : &quot;sum&quot; , payload : % Io.Eigr.Spawn.Example.MyBusinessMessage { value : 1 } ) { :ok , % Io.Eigr.Spawn.Example.MyBusinessMessage { value : 12 } } You can invoke actor default functions like &quot;get&quot; to get its current state SpawnSdk . invoke ( &quot;joe&quot; , system : &quot;spawn-system&quot; , command : &quot;get&quot; ) Spawning Actors: iex&gt; SpawnSdk . spawn_actor ( &quot;robert&quot; , system : &quot;spawn-system&quot; , actor : &quot;abs_actor&quot; ) :ok Invoke Spawned Actors: iex&gt; SpawnSdk . invoke ( &quot;robert&quot; , system : &quot;spawn-system&quot; , command : &quot;sum&quot; , payload : % Io.Eigr.Spawn.Example.MyBusinessMessage { value : 1 } ) { :ok , % Io.Eigr.Spawn.Example.MyBusinessMessage { value : 16 } } Invoke Actors in a lazy way without having to spawn them before: iex&gt; SpawnSdk . invoke ( &quot;robert_lazy&quot; , ref : SpawnSdkExample.Actors.AbstractActor , system : &quot;spawn-system&quot; , command : &quot;sum&quot; , payload : % Io.Eigr.Spawn.Example.MyBusinessMessage { value : 1 } ) { :ok , % Io.Eigr.Spawn.Example.MyBusinessMessage { value : 1 } } Invoke Actors with a delay set in milliseconds: iex&gt; SpawnSdk . invoke ( &quot;joe&quot; , system : &quot;spawn-system&quot; , command : &quot;ping&quot; , delay : 5_000 ) { :ok , :async } Invoke Actors scheduled to a specific DateTime: iex&gt; SpawnSdk . invoke ( &quot;joe&quot; , system : &quot;spawn-system&quot; , command : &quot;ping&quot; , scheduled_to : ~U[2023-01-01 00:32:00.145Z] ) { :ok , :async } Invoke Pooled Actors: iex&gt; SpawnSdk . invoke ( &quot;pooled_actor&quot; , system : &quot;spawn-system&quot; , command : &quot;ping&quot; , pooled : true ) { :ok , nil }","ref":"SpawnSdk.html","title":"SpawnSdk","type":"module"},{"doc":"Invokes a function for a actor_name Opts system this is required ref attribute attribute will always lookup to see if the referenced actor is already started or not. payload attribute is optional. command has default values that you can use to get current actor state get, get_state, Get, getState, GetState Examples iex&gt; SpawnSdk . invoke ( &quot;actor_name&quot; , ref : SpawnSdkExample.Actors.AbstractActor , system : &quot;spawn-system&quot; , command : &quot;sum&quot; , # &quot;sum&quot; or :sum payload : % Io.Eigr.Spawn.Example.MyBusinessMessage { value : 5 } ) { :ok , % Io.Eigr.Spawn.Example.MyBusinessMessage { value : 5 } } iex&gt; SpawnSdk . invoke ( &quot;actor_name&quot; , system : &quot;spawn-system&quot; , command : &quot;get&quot; ) { :ok , % Io.Eigr.Spawn.Example.MyBusinessMessage { value : 5 } }","ref":"SpawnSdk.html#invoke/2","title":"SpawnSdk.invoke/2","type":"function"},{"doc":"Spawns a abstract actor A abstract actor means that you can spawn dynamically the same actor for multiple different names. It is analog to DynamicSupervisor Opts system this is required actor which actor you will register first argument to Examples iex&gt; SpawnSdk . spawn_actor ( &quot;actor_name&quot; , system : &quot;spawn-system&quot; , actor : SpawnSdkExample.Actors.AbstractActor )","ref":"SpawnSdk.html#spawn_actor/2","title":"SpawnSdk.spawn_actor/2","type":"function"},{"doc":"Documentation for Actor . Actor look like this: defmodule MyActor do use SpawnSdk.Actor , name : &quot;joe&quot; , persistent : false , state_type : Io.Eigr.Spawn.Example.MyState , deactivate_timeout : 5_000 , snapshot_timeout : 2_000 require Logger alias Io.Eigr.Spawn.Example . { MyState , MyBusinessMessage } defact sum ( % MyBusinessMessage { value : value } = data } , % Context { state : state } = ctx ) do Logger . info ( &quot;Received Request...&quot; ) new_value = ( state . value || 0 ) + value % Value { } |&gt; Value . of ( % MyBusinessMessage { value : new_value } , % MyState { value : new_value } ) |&gt; Value . reply! ( ) end","ref":"SpawnSdk.Actor.html","title":"SpawnSdk.Actor","type":"behaviour"},{"doc":"","ref":"SpawnSdk.Actor.html#c:handle_command/2","title":"SpawnSdk.Actor.handle_command/2","type":"callback"},{"doc":"","ref":"SpawnSdk.Actor.html#t:command/0","title":"SpawnSdk.Actor.command/0","type":"type"},{"doc":"","ref":"SpawnSdk.Actor.html#t:context/0","title":"SpawnSdk.Actor.context/0","type":"type"},{"doc":"","ref":"SpawnSdk.Actor.html#t:data/0","title":"SpawnSdk.Actor.data/0","type":"type"},{"doc":"","ref":"SpawnSdk.Actor.html#t:error/0","title":"SpawnSdk.Actor.error/0","type":"type"},{"doc":"","ref":"SpawnSdk.Actor.html#t:value/0","title":"SpawnSdk.Actor.value/0","type":"type"},{"doc":"Subscriber is a helper module to subscribe Phoenix.PubSub channels","ref":"SpawnSdk.Channel.Subscriber.html","title":"SpawnSdk.Channel.Subscriber","type":"module"},{"doc":"","ref":"SpawnSdk.Channel.Subscriber.html#subscribe/2","title":"SpawnSdk.Channel.Subscriber.subscribe/2","type":"function"},{"doc":"","ref":"SpawnSdk.Channel.Subscriber.html#unsubscribe/1","title":"SpawnSdk.Channel.Subscriber.unsubscribe/1","type":"function"},{"doc":"","ref":"SpawnSdk.Channel.Subscriber.html#t:channel/0","title":"SpawnSdk.Channel.Subscriber.channel/0","type":"type"},{"doc":"","ref":"SpawnSdk.Channel.Subscriber.html#t:opts/0","title":"SpawnSdk.Channel.Subscriber.opts/0","type":"type"},{"doc":"The context is responsible for sending the State information as well as its metadata to the Actor and the Proxy and vice versa.","ref":"SpawnSdk.Context.html","title":"SpawnSdk.Context","type":"module"},{"doc":"","ref":"SpawnSdk.Context.html#t:t/0","title":"SpawnSdk.Context.t/0","type":"type"},{"doc":"Define actions like a Elixir functions Internal :defact_exports metadata saved as [ {action_name, %{timer: 10_000}}, {action_name2, %{timer: nil}} ]","ref":"SpawnSdk.Defact.html","title":"SpawnSdk.Defact","type":"module"},{"doc":"","ref":"SpawnSdk.Defact.html#defact/2","title":"SpawnSdk.Defact.defact/2","type":"macro"},{"doc":"","ref":"SpawnSdk.Flow.html","title":"SpawnSdk.Flow","type":"module"},{"doc":"Actors can also send messages to a group of actors at once as an action callback. This we call Broadcast. Example using Elixir SDK: defmodule Fleet.Actors.Driver do use SpawnSdk.Actor, kind : :abstract , # Set ´driver´ channel for all actors of the same type (Fleet.Actors.Driver) channel : &quot;drivers&quot; , state_type : Fleet.Domain.Driver alias Fleet.Domain.{ Driver , OfferRequest , OfferResponse , Point } require Logger @brain_actor_channel &quot;fleet-controllers&quot; defact update_position(%Point{} = position, %Context{state: %Driver{id: name} = driver} = ctx) do driver_state = % Driver { driver | position : position } % Value { } |&gt; Value . of ( driver_state , driver_state ) |&gt; Value . broadcast ( Broadcast . to ( @brain_actor_channel , &quot;driver_position&quot; , driver_state ) ) |&gt; Value . reply! ( ) end end In the case above, every time an Actor &quot;driver&quot; executes the update_position action it will send a message to all the actors participating in the channel called &quot;fleet-controllers&quot;. Broadcasts can also be performed outside the Spawn Actor system, using the transport mechanism based on Phoenix.PubSub in memory or Phoenix.PubSub over Nats Broker.","ref":"SpawnSdk.Flow.Broadcast.html","title":"SpawnSdk.Flow.Broadcast","type":"module"},{"doc":"","ref":"SpawnSdk.Flow.Broadcast.html#to/2","title":"SpawnSdk.Flow.Broadcast.to/2","type":"function"},{"doc":"","ref":"SpawnSdk.Flow.Broadcast.html#to/3","title":"SpawnSdk.Flow.Broadcast.to/3","type":"function"},{"doc":"","ref":"SpawnSdk.Flow.Broadcast.html#t:channel/0","title":"SpawnSdk.Flow.Broadcast.channel/0","type":"type"},{"doc":"","ref":"SpawnSdk.Flow.Broadcast.html#t:command/0","title":"SpawnSdk.Flow.Broadcast.command/0","type":"type"},{"doc":"","ref":"SpawnSdk.Flow.Broadcast.html#t:payload/0","title":"SpawnSdk.Flow.Broadcast.payload/0","type":"type"},{"doc":"","ref":"SpawnSdk.Flow.Broadcast.html#t:t/0","title":"SpawnSdk.Flow.Broadcast.t/0","type":"type"},{"doc":"Forward allows the Actor to delegate processing of the incoming message to another Actor. This is done as part of the actor's response flow. Forwards are detached from the Actor that received the input, that is, when you forward a message to another actor, the actor that performs the forwarding is free to process another message and the actor that is receiving the forwarding will respond to the original caller.","ref":"SpawnSdk.Flow.Forward.html","title":"SpawnSdk.Flow.Forward","type":"module"},{"doc":"","ref":"SpawnSdk.Flow.Forward.html#to/2","title":"SpawnSdk.Flow.Forward.to/2","type":"function"},{"doc":"","ref":"SpawnSdk.Flow.Forward.html#t:actor_name/0","title":"SpawnSdk.Flow.Forward.actor_name/0","type":"type"},{"doc":"","ref":"SpawnSdk.Flow.Forward.html#t:command/0","title":"SpawnSdk.Flow.Forward.command/0","type":"type"},{"doc":"","ref":"SpawnSdk.Flow.Forward.html#t:t/0","title":"SpawnSdk.Flow.Forward.t/0","type":"type"},{"doc":"Pipe allows the Actor to send its output message directly to another Actor, where the Actor that receives the Pipe will be responsible for following the flow from then on. This is done as part of the actor's response flow. Pipes are detached from the Actor that received the input, that is, when you forward a message to another actor through a Pipe, the actor that performs the Pipe is free to process another message and the actor that is receiving the Pipe is the one who will respond to the original caller.","ref":"SpawnSdk.Flow.Pipe.html","title":"SpawnSdk.Flow.Pipe","type":"module"},{"doc":"","ref":"SpawnSdk.Flow.Pipe.html#to/2","title":"SpawnSdk.Flow.Pipe.to/2","type":"function"},{"doc":"","ref":"SpawnSdk.Flow.Pipe.html#t:actor_name/0","title":"SpawnSdk.Flow.Pipe.actor_name/0","type":"type"},{"doc":"","ref":"SpawnSdk.Flow.Pipe.html#t:command/0","title":"SpawnSdk.Flow.Pipe.command/0","type":"type"},{"doc":"","ref":"SpawnSdk.Flow.Pipe.html#t:t/0","title":"SpawnSdk.Flow.Pipe.t/0","type":"type"},{"doc":"Actors can also emit side effects to other Actors as part of their response. Side effects do not interfere with an actor's request-response flow. They will &quot;always&quot; be processed asynchronously and any response sent back from the Actor receiving the effect will be ignored by the effector.","ref":"SpawnSdk.Flow.SideEffect.html","title":"SpawnSdk.Flow.SideEffect","type":"module"},{"doc":"","ref":"SpawnSdk.Flow.SideEffect.html#effect/5","title":"SpawnSdk.Flow.SideEffect.effect/5","type":"function"},{"doc":"","ref":"SpawnSdk.Flow.SideEffect.html#of/0","title":"SpawnSdk.Flow.SideEffect.of/0","type":"function"},{"doc":"","ref":"SpawnSdk.Flow.SideEffect.html#to/4","title":"SpawnSdk.Flow.SideEffect.to/4","type":"function"},{"doc":"","ref":"SpawnSdk.Flow.SideEffect.html#t:actor_name/0","title":"SpawnSdk.Flow.SideEffect.actor_name/0","type":"type"},{"doc":"","ref":"SpawnSdk.Flow.SideEffect.html#t:command/0","title":"SpawnSdk.Flow.SideEffect.command/0","type":"type"},{"doc":"","ref":"SpawnSdk.Flow.SideEffect.html#t:payload/0","title":"SpawnSdk.Flow.SideEffect.payload/0","type":"type"},{"doc":"","ref":"SpawnSdk.Flow.SideEffect.html#t:t/0","title":"SpawnSdk.Flow.SideEffect.t/0","type":"type"},{"doc":"Implements the communication protocol between Elixir SDK and Sidecar.","ref":"SpawnSdk.Interface.html","title":"SpawnSdk.Interface","type":"module"},{"doc":"System defines the general behavior of the Spawn actor system. It is through System implementations that the user can register, invoke, and perform other activities with their Actors.","ref":"SpawnSdk.System.html","title":"SpawnSdk.System","type":"behaviour"},{"doc":"","ref":"SpawnSdk.System.html#c:invoke/2","title":"SpawnSdk.System.invoke/2","type":"callback"},{"doc":"","ref":"SpawnSdk.System.html#c:register/2","title":"SpawnSdk.System.register/2","type":"callback"},{"doc":"","ref":"SpawnSdk.System.html#c:spawn_actor/2","title":"SpawnSdk.System.spawn_actor/2","type":"callback"},{"doc":"","ref":"SpawnSdk.System.html#t:actor_mod/0","title":"SpawnSdk.System.actor_mod/0","type":"type"},{"doc":"","ref":"SpawnSdk.System.html#t:actor_name/0","title":"SpawnSdk.System.actor_name/0","type":"type"},{"doc":"","ref":"SpawnSdk.System.html#t:actors/0","title":"SpawnSdk.System.actors/0","type":"type"},{"doc":"","ref":"SpawnSdk.System.html#t:command/0","title":"SpawnSdk.System.command/0","type":"type"},{"doc":"","ref":"SpawnSdk.System.html#t:invoke_opts/0","title":"SpawnSdk.System.invoke_opts/0","type":"type"},{"doc":"","ref":"SpawnSdk.System.html#t:payload/0","title":"SpawnSdk.System.payload/0","type":"type"},{"doc":"","ref":"SpawnSdk.System.html#t:spawn_actor_opts/0","title":"SpawnSdk.System.spawn_actor_opts/0","type":"type"},{"doc":"","ref":"SpawnSdk.System.html#t:system/0","title":"SpawnSdk.System.system/0","type":"type"},{"doc":"SpawnSystem","ref":"SpawnSdk.System.SpawnSystem.html","title":"SpawnSdk.System.SpawnSystem","type":"module"},{"doc":"","ref":"SpawnSdk.System.SpawnSystem.html#call/3","title":"SpawnSdk.System.SpawnSystem.call/3","type":"function"},{"doc":"","ref":"SpawnSdk.System.SpawnSystem.html#delete_cached_actors/2","title":"SpawnSdk.System.SpawnSystem.delete_cached_actors/2","type":"function"},{"doc":"hey","ref":"SpawnSdk.System.SpawnSystem.html#invoke/2","title":"SpawnSdk.System.SpawnSystem.invoke/2","type":"function"},{"doc":"","ref":"SpawnSdk.System.SpawnSystem.html#merge_cache_actors/2","title":"SpawnSdk.System.SpawnSystem.merge_cache_actors/2","type":"function"},{"doc":"Documentation for System.Supervisor . Start Supervisor tree like this: defmodule MyApp.Application do use Application @impl true def start ( _type , _args ) do children = [ { SpawnSdk.System.Supervisor , system : &quot;my-system&quot; , actors : [ MyActor ] } ] opts = [ strategy : :one_for_one , name : MyApp.Supervisor ] Supervisor . start_link ( children , opts ) end","ref":"SpawnSdk.System.Supervisor.html","title":"SpawnSdk.System.Supervisor","type":"module"},{"doc":"Returns a specification to start this module under a supervisor. See Supervisor .","ref":"SpawnSdk.System.Supervisor.html#child_spec/1","title":"SpawnSdk.System.Supervisor.child_spec/1","type":"function"},{"doc":"","ref":"SpawnSdk.System.Supervisor.html#start_link/1","title":"SpawnSdk.System.Supervisor.start_link/1","type":"function"},{"doc":"Value is a declarative DSL that provides the Domain Driven aspect of the Spawn technology. It is through a Value that the user can configure the proxy to carry out the infrastructure tasks and from which it provides all the Worflows.","ref":"SpawnSdk.Value.html","title":"SpawnSdk.Value","type":"module"},{"doc":"","ref":"SpawnSdk.Value.html#broadcast/2","title":"SpawnSdk.Value.broadcast/2","type":"function"},{"doc":"","ref":"SpawnSdk.Value.html#effect/2","title":"SpawnSdk.Value.effect/2","type":"function"},{"doc":"","ref":"SpawnSdk.Value.html#effects/2","title":"SpawnSdk.Value.effects/2","type":"function"},{"doc":"","ref":"SpawnSdk.Value.html#forward/2","title":"SpawnSdk.Value.forward/2","type":"function"},{"doc":"","ref":"SpawnSdk.Value.html#noreply!/2","title":"SpawnSdk.Value.noreply!/2","type":"function"},{"doc":"","ref":"SpawnSdk.Value.html#noreply_state!/1","title":"SpawnSdk.Value.noreply_state!/1","type":"function"},{"doc":"","ref":"SpawnSdk.Value.html#of/0","title":"SpawnSdk.Value.of/0","type":"function"},{"doc":"","ref":"SpawnSdk.Value.html#of/2","title":"SpawnSdk.Value.of/2","type":"function"},{"doc":"","ref":"SpawnSdk.Value.html#of/3","title":"SpawnSdk.Value.of/3","type":"function"},{"doc":"","ref":"SpawnSdk.Value.html#pipe/2","title":"SpawnSdk.Value.pipe/2","type":"function"},{"doc":"","ref":"SpawnSdk.Value.html#reply!/1","title":"SpawnSdk.Value.reply!/1","type":"function"},{"doc":"","ref":"SpawnSdk.Value.html#response/2","title":"SpawnSdk.Value.response/2","type":"function"},{"doc":"","ref":"SpawnSdk.Value.html#state/2","title":"SpawnSdk.Value.state/2","type":"function"},{"doc":"","ref":"SpawnSdk.Value.html#tags/2","title":"SpawnSdk.Value.tags/2","type":"function"},{"doc":"","ref":"SpawnSdk.Value.html#value/2","title":"SpawnSdk.Value.value/2","type":"function"},{"doc":"","ref":"SpawnSdk.Value.html#void/0","title":"SpawnSdk.Value.void/0","type":"function"},{"doc":"","ref":"SpawnSdk.Value.html#void/1","title":"SpawnSdk.Value.void/1","type":"function"},{"doc":"","ref":"SpawnSdk.Value.html#t:broadcast/0","title":"SpawnSdk.Value.broadcast/0","type":"type"},{"doc":"","ref":"SpawnSdk.Value.html#t:effects/0","title":"SpawnSdk.Value.effects/0","type":"type"},{"doc":"","ref":"SpawnSdk.Value.html#t:forward/0","title":"SpawnSdk.Value.forward/0","type":"type"},{"doc":"","ref":"SpawnSdk.Value.html#t:new_state/0","title":"SpawnSdk.Value.new_state/0","type":"type"},{"doc":"","ref":"SpawnSdk.Value.html#t:pipe/0","title":"SpawnSdk.Value.pipe/0","type":"type"},{"doc":"","ref":"SpawnSdk.Value.html#t:response/0","title":"SpawnSdk.Value.response/0","type":"type"},{"doc":"","ref":"SpawnSdk.Value.html#t:t/0","title":"SpawnSdk.Value.t/0","type":"type"},{"doc":"","ref":"SpawnSdk.Value.html#t:tags/0","title":"SpawnSdk.Value.tags/0","type":"type"},{"doc":"","ref":"SpawnSdk.Value.html#t:value/0","title":"SpawnSdk.Value.value/0","type":"type"},{"doc":"Spawn Elixir SDK is the support library for the Spawn Actors System. Spawn is a Stateful Serverless Platform for providing the multi-language Actor Model. For a broader understanding of Spawn, please consult its official repository . The advantage of the Elixir SDK over other SDKs is in Elixir's native ability to connect directly to an Erlang network. For this reason, the Elixir SDK allows any valid Elixir application to be part of a Spawn network without needing a sidecar attached.","ref":"readme.html","title":"Spawn Elixir SDK","type":"extras"},{"doc":"Available in Hex , the package can be installed by adding spawn_sdk and spawn_statestores_* to your list of dependencies in mix.exs : def deps do [ { :spawn_sdk , &quot;~&gt; 1.0.0-rc1&quot; } , # You can uncomment one of those dependencies if you are going to use Persistent Actors \#{:spawn_statestores_mysql, &quot;~&gt; 1.0.0-rc1&quot;}, \#{:spawn_statestores_postgres, &quot;~&gt; 1.0.0-rc1&quot;}, \#{:spawn_statestores_mssql, &quot;~&gt; 1.0.0-rc1&quot;}, \#{:spawn_statestores_cockroachdb, &quot;~&gt; 1.0.0-rc1&quot;}, \#{:spawn_statestores_sqlite, &quot;~&gt; 1.0.0-rc1&quot;}, ] end Deploy Following the steps below, you will have a valid Elixir application to use in a Spawn cluster. However, you will still need to generate a container image with your application to use it together with the Spawn Operator for Kubernetes .","ref":"readme.html#installation","title":"Spawn Elixir SDK - Installation","type":"extras"},{"doc":"After creating an Elixir application project, create the protobuf files for your business domain. It is common practice to do this under the priv/ folder. Let's demonstrate an example: syntax = &quot;proto3&quot;; package io.eigr.spawn.example; message MyState { int32 value = 1; } message MyBusinessMessage { int32 value = 1; } It is important to try to separate the type of message that must be stored as the actors' state from the type of messages that will be exchanged between their actors' operations calls. In other words, the Actor's internal state is also represented as a protobuf type, and it is a good practice to separate these types of messages from the others in its business domain. In the above case MyState is the type protobuf that represents the state of the Actor that we will create later while MyBusiness is the type of message that we will send and receive from this Actor. Now that we have defined our input and output types as Protobuf types we will need to compile these files to generate their respective Elixir modules. An example of how to do this can be found here NOTE: You need to have installed the elixir plugin for protoc. More information on how to obtain and install the necessary tools can be found here here Now that the protobuf types have been created we can proceed with the code. Example definition of an Actor.","ref":"readme.html#how-to-use","title":"Spawn Elixir SDK - How to use","type":"extras"},{"doc":"In this example we are creating an actor in a Singleton way, that is, it is a known actor at compile time. defmodule SpawnSdkExample.Actors.MyActor do use SpawnSdk.Actor , name : &quot;jose&quot; , # Default is Full Qualified Module name a.k.a __MODULE__ kind : :singleton , # Default is already :singleton. Valid are :singleton | :abstract | :pooled stateful : true , # Default is already true state_type : Io.Eigr.Spawn.Example.MyState , # or :json if you don&#39;t care about protobuf types deactivate_timeout : 30_000 , snapshot_timeout : 2_000 require Logger alias Io.Eigr.Spawn.Example . { MyState , MyBusinessMessage } defact init ( % Context { state : state } = ctx ) do Logger . info ( &quot;[joe] Received InitRequest. Context: \#{ inspect ( ctx ) } &quot; ) Value . of ( ) |&gt; Value . state ( state ) end defact sum ( % MyBusinessMessage { value : value } = data , % Context { state : state } = ctx ) do Logger . info ( &quot;Received Request: \#{ inspect ( data ) } . Context: \#{ inspect ( ctx ) } &quot; ) new_value = if is_nil ( state ) , do : value , else : ( state . value || 0 ) + value Value . of ( % MyBusinessMessage { value : new_value } , % MyState { value : new_value } ) end end We declare two actions that the Actor can do. An initialization action that will be called every time an Actor instance is created and an action that will be responsible for performing a simple sum. Note Keep in mind that any Action that has the names present in the list below will behave as an initialization Action and will be called when the Actor is started (if there is more than one Action with one of these names, only one will be called). Defaults inicialization Action names: &quot; init &quot;, &quot; Init &quot;, &quot; setup &quot;, &quot; Setup &quot;","ref":"readme.html#singleton-actors","title":"Spawn Elixir SDK - Singleton Actors","type":"extras"},{"doc":"We can also create Unnamed Dynamic/Lazy actors, that is, despite having its abstract behavior defined at compile time, a Lazy actor will only have a concrete instance when it is associated with an identifier/name at runtime. Below follows the same previous actor being defined as abstract. defmodule SpawnSdkExample.Actors.AbstractActor do use SpawnSdk.Actor , name : &quot;abs_actor&quot; , kind : :abstract , state_type : Io.Eigr.Spawn.Example.MyState require Logger alias Io.Eigr.Spawn.Example . { MyState , MyBusinessMessage } defact sum ( % MyBusinessMessage { value : value } = data , % Context { state : state } = ctx ) do Logger . info ( &quot;Received Request: \#{ inspect ( data ) } . Context: \#{ inspect ( ctx ) } &quot; ) new_value = if is_nil ( state ) , do : value , else : ( state . value || 0 ) + value Value . of ( % MyBusinessMessage { value : new_value } , % MyState { value : new_value } ) end end Notice that the only thing that has changed is the the kind of actor, in this case the kind is set to :abstract. NOTE: Can Elixir programmers think in terms of singleton vs abstract actors as more or less known at startup vs dynamically supervised/registered? That is, defining your actors directly in the supervision tree or using a Dynamic Supervisor for that.","ref":"readme.html#abstract-actor","title":"Spawn Elixir SDK - Abstract Actor","type":"extras"},{"doc":"Sometimes we want a particular actor to be able to serve requests concurrently, however actors will always serve one request at a time using buffering mechanisms to receive requests in their mailbox and serve each request one by one. So to get around this behaviour you can configure your Actor as a Pooled Actor, this way the system will generate a pool of actors to meet certain requests. See an example below: defmodule SpawnSdkExample.Actors.PooledActor do use SpawnSdk.Actor , name : &quot;pooled_actor&quot; , kind : :pooled , stateful : false require Logger defact ping ( _data , % Context { } = ctx ) do Logger . info ( &quot;Received Request. Context: \#{ inspect ( ctx ) } &quot; ) Value . void ( ) end end NOTE: It is important to mention that Pooled Actors cannot be stateful because concurrent accesses for Stateful actors could generate undesired inconsistencies. To invoke a pooled actor, it is necessary to pass the pooled: true option to the invoke function. See an example: iex ( spawn_a1 @ 127.0 . 0.1 ) 16 &gt; for _n &lt;- 0 .. 8 , do : SpawnSdk . invoke ( &quot;pooled_actor&quot; , system : &quot;spawn-system&quot; , command : &quot;ping&quot; , pooled : true ) In the logs you will see that even though you have invoked the actor by the registered name more than one instance of the actor will be created. As we can see below: iex ( spawn_a1 @ 127.0 . 0.1 ) 16 &gt; for _n &lt;- 0 .. 8 , do : SpawnSdk . invoke ( &quot;pooled_actor&quot; , system : &quot;spawn-system&quot; , command : &quot;ping&quot; , pooled : true ) 2022 - 11 - 25 11 : 21 : 53.758 [ spawn_a1 @ 127.0 . 0.1 ] : [ pid = &lt; 0.1392 . 0 &gt; ] : [ notice ] :Activating Actor pooled_actor - 5 with Parent pooled_actor - 5 in Node :&quot;spawn_a1@127.0.0.1&quot; . Persistence false . 2022 - 11 - 25 11 : 21 : 53.759 [ spawn_a1 @ 127.0 . 0.1 ] : [ pid = &lt; 0.1391 . 0 &gt; ] : [ debug ] :Actor pooled_actor - 5 reactivated . ActorRef PID : # PID &lt; 0 . 1392 . 0 &gt; 2022 - 11 - 25 11 : 21 : 53.760 [ spawn_a1 @ 127.0 . 0.1 ] : [ pid = &lt; 0.1392 . 0 &gt; ] : [ info ] :Received Request . Context : % SpawnSdk.Context { state : nil , caller : nil , self : % Eigr.Functions.Protocol.Actors.ActorId { name : &quot;pooled_actor-5&quot; , system : &quot;spawn-system&quot; , parent : &quot;pooled_actor-5&quot; , __unknown_fields__ : [ ] } , metadata : %{ } } 2022 - 11 - 25 11 : 21 : 53.763 [ spawn_a1 @ 127.0 . 0.1 ] : [ pid = &lt; 0.1394 . 0 &gt; ] : [ notice ] :Activating Actor pooled_actor - 4 with Parent pooled_actor - 4 in Node :&quot;spawn_a1@127.0.0.1&quot; . Persistence false . 2022 - 11 - 25 11 : 21 : 53.763 [ spawn_a1 @ 127.0 . 0.1 ] : [ pid = &lt; 0.1393 . 0 &gt; ] : [ debug ] :Actor pooled_actor - 4 reactivated . ActorRef PID : # PID &lt; 0 . 1394 . 0 &gt; 2022 - 11 - 25 11 : 21 : 53.765 [ spawn_a1 @ 127.0 . 0.1 ] : [ pid = &lt; 0.1394 . 0 &gt; ] : [ info ] :Received Request . Context : % SpawnSdk.Context { state : nil , caller : nil , self : % Eigr.Functions.Protocol.Actors.ActorId { name : &quot;pooled_actor-4&quot; , system : &quot;spawn-system&quot; , parent : &quot;pooled_actor-4&quot; , __unknown_fields__ : [ ] } , metadata : %{ } } 2022 - 11 - 25 11 : 21 : 53.767 [ spawn_a1 @ 127.0 . 0.1 ] : [ pid = &lt; 0.1396 . 0 &gt; ] : [ notice ] :Activating Actor pooled_actor - 7 with Parent pooled_actor - 7 in Node :&quot;spawn_a1@127.0.0.1&quot; . Persistence false . 2022 - 11 - 25 11 : 21 : 53.767 [ spawn_a1 @ 127.0 . 0.1 ] : [ pid = &lt; 0.1395 . 0 &gt; ] : [ debug ] :Actor pooled_actor - 7 reactivated . ActorRef PID : # PID &lt; 0 . 1396 . 0 &gt; 2022 - 11 - 25 11 : 21 : 53.768 [ spawn_a1 @ 127.0 . 0.1 ] : [ pid = &lt; 0.1396 . 0 &gt; ] : [ info ] :Received Request . Context : % SpawnSdk.Context { state : nil , caller : nil , self : % Eigr.Functions.Protocol.Actors.ActorId { name : &quot;pooled_actor-7&quot; , system : &quot;spawn-system&quot; , parent : &quot;pooled_actor-7&quot; , __unknown_fields__ : [ ] } , metadata : %{ } } 2022 - 11 - 25 11 : 21 : 53.770 [ spawn_a1 @ 127.0 . 0.1 ] : [ pid = &lt; 0.1398 . 0 &gt; ] : [ notice ] :Activating Actor pooled_actor - 2 with Parent pooled_actor - 2 in Node :&quot;spawn_a1@127.0.0.1&quot; . Persistence false . 2022 - 11 - 25 11 : 21 : 53.771 [ spawn_a1 @ 127.0 . 0.1 ] : [ pid = &lt; 0.1397 . 0 &gt; ] : [ debug ] :Actor pooled_actor - 2 reactivated . ActorRef PID : # PID &lt; 0 . 1398 . 0 &gt; 2022 - 11 - 25 11 : 21 : 53.772 [ spawn_a1 @ 127.0 . 0.1 ] : [ pid = &lt; 0.1398 . 0 &gt; ] : [ info ] :Received Request . Context : % SpawnSdk.Context { state : nil , caller : nil , self : % Eigr.Functions.Protocol.Actors.ActorId { name : &quot;pooled_actor-2&quot; , system : &quot;spawn-system&quot; , parent : &quot;pooled_actor-2&quot; , __unknown_fields__ : [ ] } , metadata : %{ } } 2022 - 11 - 25 11 : 21 : 53.774 [ spawn_a1 @ 127.0 . 0.1 ] : [ pid = &lt; 0.1400 . 0 &gt; ] : [ notice ] :Activating Actor pooled_actor - 8 with Parent pooled_actor - 8 in Node :&quot;spawn_a1@127.0.0.1&quot; . Persistence false . 2022 - 11 - 25 11 : 21 : 53.775 [ spawn_a1 @ 127.0 . 0.1 ] : [ pid = &lt; 0.1399 . 0 &gt; ] : [ debug ] :Actor pooled_actor - 8 reactivated . ActorRef PID : # PID &lt; 0 . 1400 . 0 &gt; 2022 - 11 - 25 11 : 21 : 53.776 [ spawn_a1 @ 127.0 . 0.1 ] : [ pid = &lt; 0.1400 . 0 &gt; ] : [ info ] :Received Request . Context : % SpawnSdk.Context { state : nil , caller : nil , self : % Eigr.Functions.Protocol.Actors.ActorId { name : &quot;pooled_actor-8&quot; , system : &quot;spawn-system&quot; , parent : &quot;pooled_actor-8&quot; , __unknown_fields__ : [ ] } , metadata : %{ } } 2022 - 11 - 25 11 : 21 : 53.777 [ spawn_a1 @ 127.0 . 0.1 ] : [ pid = &lt; 0.1402 . 0 &gt; ] : [ notice ] :Activating Actor pooled_actor - 9 with Parent pooled_actor - 9 in Node :&quot;spawn_a1@127.0.0.1&quot; . Persistence false . 2022 - 11 - 25 11 : 21 : 53.778 [ spawn_a1 @ 127.0 . 0.1 ] : [ pid = &lt; 0.1401 . 0 &gt; ] : [ debug ] :Actor pooled_actor - 9 reactivated . ActorRef PID : # PID &lt; 0 . 1402 . 0 &gt; 2022 - 11 - 25 11 : 21 : 53.778 [ spawn_a1 @ 127.0 . 0.1 ] : [ pid = &lt; 0.1402 . 0 &gt; ] : [ info ] :Received Request . Context : % SpawnSdk.Context { state : nil , caller : nil , self : % Eigr.Functions.Protocol.Actors.ActorId { name : &quot;pooled_actor-9&quot; , system : &quot;spawn-system&quot; , parent : &quot;pooled_actor-9&quot; , __unknown_fields__ : [ ] } , metadata : %{ } } 2022 - 11 - 25 11 : 21 : 53.779 [ spawn_a1 @ 127.0 . 0.1 ] : [ pid = &lt; 0.1212 . 0 &gt; ] : [ debug ] :Lookup Actor pooled_actor . PID : # PID &lt; 0 . 1400 . 0 &gt; 2022 - 11 - 25 11 : 21 : 53.780 [ spawn_a1 @ 127.0 . 0.1 ] : [ pid = &lt; 0.1400 . 0 &gt; ] : [ info ] :Received Request . Context : % SpawnSdk.Context { state : nil , caller : nil , self : % Eigr.Functions.Protocol.Actors.ActorId { name : &quot;pooled_actor-8&quot; , system : &quot;spawn-system&quot; , parent : &quot;pooled_actor-8&quot; , __unknown_fields__ : [ ] } , metadata : %{ } } 2022 - 11 - 25 11 : 21 : 53.780 [ spawn_a1 @ 127.0 . 0.1 ] : [ pid = &lt; 0.1212 . 0 &gt; ] : [ debug ] :Lookup Actor pooled_actor . PID : # PID &lt; 0 . 1392 . 0 &gt; 2022 - 11 - 25 11 : 21 : 53.781 [ spawn_a1 @ 127.0 . 0.1 ] : [ pid = &lt; 0.1392 . 0 &gt; ] : [ info ] :Received Request . Context : % SpawnSdk.Context { state : nil , caller : nil , self : % Eigr.Functions.Protocol.Actors.ActorId { name : &quot;pooled_actor-5&quot; , system : &quot;spawn-system&quot; , parent : &quot;pooled_actor-5&quot; , __unknown_fields__ : [ ] } , metadata : %{ } } 2022 - 11 - 25 11 : 21 : 53.781 [ spawn_a1 @ 127.0 . 0.1 ] : [ pid = &lt; 0.1212 . 0 &gt; ] : [ debug ] :Lookup Actor pooled_actor . PID : # PID &lt; 0 . 1394 . 0 &gt; [ ok : nil , ok : nil , ok : nil , ok : nil , ok : nil , ok : nil , ok : nil , ok : nil , ok : nil ] iex ( spawn_a1 @ 127.0 . 0.1 ) 17 &gt; 2022 - 11 - 25 11 : 21 : 53.782 [ spawn_a1 @ 127.0 . 0.1 ] : [ pid = &lt; 0.1394 . 0 &gt; ] : [ info ] :Received Request . Context : % SpawnSdk.Context { state : nil , caller : nil , self : % Eigr.Functions.Protocol.Actors.ActorId { name : &quot;pooled_actor-4&quot; , system : &quot;spawn-system&quot; , parent : &quot;pooled_actor-4&quot; , __unknown_fields__ : [ ] } , metadata : %{ } } 2022 - 11 - 25 11 : 22 : 03.999 [ spawn_a1 @ 127.0 . 0.1 ] : [ pid = &lt; 0.1402 . 0 &gt; ] : [ debug ] :Deactivating actor pooled_actor - 9 for timeout 2022 - 11 - 25 11 : 22 : 04.226 [ spawn_a1 @ 127.0 . 0.1 ] : [ pid = &lt; 0.1392 . 0 &gt; ] : [ debug ] :Deactivating actor pooled_actor - 5 for timeout 2022 - 11 - 25 11 : 22 : 07.271 [ spawn_a1 @ 127.0 . 0.1 ] : [ pid = &lt; 0.1400 . 0 &gt; ] : [ debug ] :Deactivating actor pooled_actor - 8 for timeout 2022 - 11 - 25 11 : 22 : 10.167 [ spawn_a1 @ 127.0 . 0.1 ] : [ pid = &lt; 0.1396 . 0 &gt; ] : [ debug ] :Deactivating actor pooled_actor - 7 for timeout 2022 - 11 - 25 11 : 22 : 12.038 [ spawn_a1 @ 127.0 . 0.1 ] : [ pid = &lt; 0.1394 . 0 &gt; ] : [ debug ] :Deactivating actor pooled_actor - 4 for timeout 2022 - 11 - 25 11 : 22 : 12.613 [ spawn_a1 @ 127.0 . 0.1 ] : [ pid = &lt; 0.1398 . 0 &gt; ] : [ debug ] :Deactivating actor pooled_actor - 2 for timeout It is possible to configure the minimum and maximum pool size by passing the min_pool_size and max_pool_size options. The actor below would create a pool of a maximum of 100 Actors: defmodule SpawnSdkExample.Actors.PooledActor do use SpawnSdk.Actor , name : &quot;pooled_actor&quot; , kind : :pooled , min_pool_size : 1 , max_pool_size : 10 , stateful : false # Code omitted for brevity end ALERT: Keep in mind that a large pool size does not always mean better performance. In fact, smaller pool sizes sometimes often have a much greater effect in this regard. It is interesting that you experiment with different pool sizes until you find the one that best suits your workload.","ref":"readme.html#pooled-actors","title":"Spawn Elixir SDK - Pooled Actors","type":"extras"},{"doc":"Actors can also emit side effects to other Actors as part of their response. See an example: defmodule SpawnSdkExample.Actors.AbstractActor do use SpawnSdk.Actor , kind : :abstract , stateful : false , state_type : Io.Eigr.Spawn.Example.MyState require Logger alias Io.Eigr.Spawn.Example . { MyState , MyBusinessMessage } alias SpawnSdk.Flow.SideEffect defact sum ( % MyBusinessMessage { value : value } = data , % Context { state : state } = ctx ) do Logger . info ( &quot;Received Request: \#{ inspect ( data ) } . Context: \#{ inspect ( ctx ) } &quot; ) new_value = if is_nil ( state ) , do : value , else : ( state . value || 0 ) + value result = % MyBusinessMessage { value : new_value } new_state = % MyState { value : new_value } Value . of ( ) |&gt; Value . response ( result ) |&gt; Value . state ( new_state ) |&gt; Value . effects ( # This returns a list of side effects. In this case containing only one effect. However, multiple effects can be chained together, # just by calling the effect function as shown here. # The delay means that it will be fired asynchronously after 5000 milliseconds (5 seconds) # If only one effect is desired, you can also choose to use the to/3 function together with Value.effect(). # Example: Values.effect(SideEffect.to(name, func, payload)) SideEffect . of ( ) |&gt; SideEffect . effect ( &quot;joe&quot; , :sum , result , delay : 5_000 , scheduled_to : ~U[2020-01-01 10:00:00.145Z] ) # use delay or scheduled_to, not both ) end end In the example above we see that the Actor joe will receive a request as a side effect from the Actor who issued this effect. Side effects do not interfere with an actor's request-response flow. They will &quot;always&quot; be processed asynchronously and any response sent back from the Actor receiving the effect will be ignored by the effector.","ref":"readme.html#side-effects","title":"Spawn Elixir SDK - Side Effects","type":"extras"},{"doc":"Actors can also route some commands to other actors as part of their response. See an example: defmodule SpawnSdkExample.Actors.ForwardPipeActor do use SpawnSdk.Actor , name : &quot;pipeforward&quot; , kind : :singleton , stateful : false require Logger alias Io.Eigr.Spawn.Example.MyBusinessMessage defact forward_example ( % MyBusinessMessage { } = msg , _ctx ) do Logger . info ( &quot;Received request with \#{ msg . value } &quot; ) Value . of ( ) |&gt; Value . response ( MyBusinessMessage . new ( value : 999 ) ) |&gt; Value . forward ( Forward . to ( &quot;second_actor&quot; , &quot;sum_plus_one&quot; ) ) |&gt; Value . void ( ) end defact pipe_example ( % MyBusinessMessage { } = msg , _ctx ) do Logger . info ( &quot;Received request with \#{ msg . value } &quot; ) Value . of ( ) |&gt; Value . response ( MyBusinessMessage . new ( value : 999 ) ) |&gt; Value . pipe ( Pipe . to ( &quot;second_actor&quot; , &quot;sum_plus_one&quot; ) ) |&gt; Value . void ( ) end end defmodule SpawnSdkExample.Actors.SecondActorExample do use SpawnSdk.Actor , name : &quot;second_actor&quot; , stateful : false require Logger alias Io.Eigr.Spawn.Example.MyBusinessMessage defact sum_plus_one ( % MyBusinessMessage { } = msg , _ctx ) do Logger . info ( &quot;Received request with \#{ msg . value } &quot; ) Value . of ( ) |&gt; Value . response ( MyBusinessMessage . new ( value : msg . value + 1 ) ) |&gt; Value . void ( ) end end We are returning void in both examples so we dont care about what is being stored in the actor state. In the case above, every time you call the forward_example the second_actor's sum_plus_one function will receive the value forwarded originally in the invocation as its input. The end result will be: iex&gt; SpawnSdk . invoke ( &quot;pipeforward&quot; , system : &quot;spawn-system&quot; , command : &quot;forward_example&quot; , payload : % Io.Eigr.Spawn.Example.MyBusinessMessage { value : 1 } ) { :ok , % Io.Eigr.Spawn.Example.MyBusinessMessage { value : 2 } } For the Pipe example, the the second_actor's sum_plus_one function will always receive %MyBusinessMessage{value: 999} due to getting the value from the previous specification in the pipe_example command, the end result will be: iex&gt; SpawnSdk . invoke ( &quot;pipeforward&quot; , system : &quot;spawn-system&quot; , command : &quot;pipe_example&quot; , payload : % Io.Eigr.Spawn.Example.MyBusinessMessage { value : 1 } ) { :ok , % Io.Eigr.Spawn.Example.MyBusinessMessage { value : 1000 } }","ref":"readme.html#pipe-and-forward","title":"Spawn Elixir SDK - Pipe and Forward","type":"extras"},{"doc":"Actors can also send messages to a group of actors at once as an action callback. See the example below: defmodule Fleet.Actors.Driver do use SpawnSdk.Actor , kind : :abstract , # Set ´driver´ channel for all actors of the same type (Fleet.Actors.Driver) channel : &quot;drivers&quot; , state_type : Fleet.Domain.Driver alias Fleet.Domain . { Driver , OfferRequest , OfferResponse , Point } require Logger @brain_actor_channel &quot;fleet-controllers&quot; defact update_position ( % Point { } = position , % Context { state : % Driver { id : name } = driver } = ctx ) do Logger . info ( &quot;Driver [ \#{ name } ] Received Update Position Event. Position: [ \#{ inspect ( position ) } ]. Context: \#{ inspect ( ctx ) } &quot; ) driver_state = % Driver { driver | position : position } % Value { } |&gt; Value . of ( driver_state , driver_state ) |&gt; Value . broadcast ( Broadcast . to ( @brain_actor_channel , &quot;driver_position&quot; , driver_state ) ) end end In the case above, every time an Actor &quot;driver&quot; executes the update_position action it will send a message to all the actors participating in the channel called &quot;fleet-controllers&quot;. Broadcast to External Subscribers Sometimes you may want to send events out of ActorSystem using Phoenix.PubSub. One way to do this is to take advantage of the same Broadcast infrastructure that Spawn offers you but indicating an external channel. Below is an example: Create a Listener to receive the events using the SpawnSdk.Channel.Subscriber helper module. defmodule SpawnSdkExample.Subscriber do @moduledoc &quot;&quot;&quot; This module exemplifies how to listen for pubsub events that were emitted by actors but that will be treated not by actors but as normal pubsub events. This is particularly useful for integrations between Spawn and Phoenix LiveView. &quot;&quot;&quot; use GenServer require logger alias SpawnSdk.Channel.Subscriber @impl true define init ( state ) do Subscriber . subscribe ( &quot;external.channel&quot; ) { :ok , state } end @impl true def handle_info ( { :receive , payload } , state ) do Logger . info ( &quot;Received pubsub event \#{ inspect ( payload ) } &quot; ) { :noreply , state } end def start_link ( args ) do GenServer . start_link ( __MODULE__ , args ) end end You need to match using the {:receive, payload} tuple in your handle_info. NOTE: By default SpawnSdk.Channel.Subscriber will subscribe to pubsub using the atom :actor_channel as an argument. If you need to change this , just configure your configuration as follows : config.exs config :spawn , pubsub_group : :your_channel_group_here SpawnSdk.System.Supervisor. defmodule SpawnSdkExample.Application do @moduledoc false use Application @impl true def start ( _type , _args ) do children = [ { SpawnSdk.System.Supervisor , system : &quot;spawn-system&quot; , actors : [ SpawnSdkExample.Actors.JoeActor ] , extenal_subscribers : [ { SpawnSdkExample.Subscriber , [ ] } ] } ] opts = [ strategy : :one_for_one , name : SpawnSdkExample.Supervisor ] Supervisor . start_link ( children , opts ) end end The important thing here is to use the external_subscribers attribute. As seen above :external_subscribers accepts a list of specs as a parameter. Set your actor as you normally would and emit your broadcast events using Broadcast.to(channel, payload). defmodule SpawnSdkExample.Actors.JoeActor do use SpawnSdk.Actor , name : &quot;joe&quot; , state_type : Io.Eigr.Spawn.Example.MyState require Logger alias Io.Eigr.Spawn.Example . { MyState , MyBusinessMessage } defact sum ( % MyBusinessMessage { value : value } = data , % Context { state : state } = ctx ) do Logger . info ( &quot;[joe] Received Request: \#{ inspect ( data ) } . Context: \#{ inspect ( ctx ) } &quot; ) new_value = if is_nil ( state ) do 0 + value else ( state . value || 0 ) + value end response = % MyBusinessMessage { value : new_value } % Value { } |&gt; Value . of ( response , % MyState { value : new_value } ) |&gt; Value . broadcast ( Broadcast . to ( &quot;my.channel&quot; , response ) ) end end","ref":"readme.html#broadcast","title":"Spawn Elixir SDK - Broadcast","type":"extras"},{"doc":"Actors can also declare Actions that act recursively as timers. See an example below: defmodule SpawnSdkExample.Actors.ClockActor do use SpawnSdk.Actor , name : &quot;clock_actor&quot; , state_type : Io.Eigr.Spawn.Example.MyState , deactivate_timeout : 86_400_000 require Logger alias Io.Eigr.Spawn.Example.MyState @set_timer 15_000 defact clock ( % Context { state : state } = ctx ) do Logger . info ( &quot;[clock] Clock Actor Received Request. Context: \#{ inspect ( ctx ) } &quot; ) new_value = if is_nil ( state ) , do : 0 , else : state . value + 1 new_state = MyState . new ( value : new_value ) Value . of ( ) |&gt; Value . state ( new_state ) end end NOTE: Timers Actions are recorded as Actor metadata. Where in turn we use a synchronization mechanism via CRDTs to keep the metadata alive in the cluster while there is an active Spawn VM. That is, Timers Actions are ephemeral and therefore only exist while there is at least one active VM in the cluster. In the example above the ´clock´ action will be called every 15 seconds.","ref":"readme.html#timers","title":"Spawn Elixir SDK - Timers","type":"extras"},{"doc":"Once we define our actors we can now declare our supervision tree: defmodule SpawnSdkExample.Application do @moduledoc false use Application @impl true def start ( _type , _args ) do children = [ { SpawnSdk.System.Supervisor , system : &quot;spawn-system&quot; , actors : [ SpawnSdkExample.Actors.MyActor , SpawnSdkExample.Actors.AbstractActor , SpawnSdkExample.Actors.ClockActor , SpawnSdkExample.Actors.PooledActor ] } ] opts = [ strategy : :one_for_one , name : SpawnSdkExample.Supervisor ] Supervisor . start_link ( children , opts ) end end","ref":"readme.html#declaring-the-supervision-tree","title":"Spawn Elixir SDK - Declaring the supervision tree","type":"extras"},{"doc":"Actors also have some standard actions that are not implemented by the user and that can be used as a way to get the state of an actor without the invocation requiring an extra trip to the host functions. You can think of them as a cache of their state, every time you invoke a default action on an actor it will return the value directly from the Sidecar process without this target process needing to invoke its equivalent host function. Let's take an example. Suppose Actor Joe wants to know the current state of Actor Robert. What Joe can do is invoke Actor Robert's default action called get_state. This will make Actor Joe's sidecar find Actor Robert's sidecar somewhere in the cluster and Actor Robert's sidecar will return its own state directly to Joe without having to resort to your host function, this in turn will save you a called over the network and therefore this type of invocation is faster than invocations of user-defined actions usually are. Any invocations to actions with the following names will follow this rule: &quot; get &quot;, &quot; Get &quot;, &quot; get_state &quot;, &quot; getState &quot;, &quot; GetState &quot; NOTE: You can override this behavior by defining your actor as an action with the same name as the default actions. In this case it will be the Action defined by you that will be called, implying perhaps another network roundtrip","ref":"readme.html#default-actions","title":"Spawn Elixir SDK - Default Actions","type":"extras"},{"doc":"To deploy your actors in a cluster you need to use our Controller for Kubernetes. Detailed information on how to proceed can be found here and here . But you can also run your Elixir application in the traditional way as follows: MIX_ENV = prod USER_FUNCTION_PORT = 8092 PROXY_DATABASE_TYPE = mysql SPAWN_STATESTORE_KEY = 3 Jnb0hZiHIzHTOih7t2cTEPEpY98Tu1wvQkPfq / XwqE = iex -- name spawn_a2 @ 127.0 . 0.1 - S mix NOTE: For reasons of compatibility with our controller, it is necessary to configure your Spawn application using environment variables instead of the traditional Elixir configuration mechanism. WARNING: This example uses the MySQL database as persistent storage for its actors. And it is also expected that you have previously created a database called eigr-functions-db in the MySQL instance. The full example of this application can be found here . And links to other examples can be found in our github readme page .","ref":"readme.html#running","title":"Spawn Elixir SDK - Running","type":"extras"},{"doc":"To invoke Actors, use: iex&gt; SpawnSdk . invoke ( &quot;joe&quot; , system : &quot;spawn-system&quot; , command : &quot;sum&quot; , payload : % Io.Eigr.Spawn.Example.MyBusinessMessage { value : 1 } ) { :ok , % Io.Eigr.Spawn.Example.MyBusinessMessage { value : 12 } } You can invoke actor default functions like &quot;get&quot; to get its current state SpawnSdk . invoke ( &quot;joe&quot; , system : &quot;spawn-system&quot; , command : &quot;get&quot; ) Spawning Actors: iex&gt; SpawnSdk . spawn_actor ( &quot;robert&quot; , system : &quot;spawn-system&quot; , actor : &quot;abs_actor&quot; ) :ok Invoke Spawned Actors: iex&gt; SpawnSdk . invoke ( &quot;robert&quot; , system : &quot;spawn-system&quot; , command : &quot;sum&quot; , payload : % Io.Eigr.Spawn.Example.MyBusinessMessage { value : 1 } ) { :ok , % Io.Eigr.Spawn.Example.MyBusinessMessage { value : 16 } } Invoke Actors in a lazy way without having to spawn them before: iex&gt; SpawnSdk . invoke ( &quot;robert_lazy&quot; , ref : SpawnSdkExample.Actors.AbstractActor , system : &quot;spawn-system&quot; , command : &quot;sum&quot; , payload : % Io.Eigr.Spawn.Example.MyBusinessMessage { value : 1 } ) { :ok , % Io.Eigr.Spawn.Example.MyBusinessMessage { value : 1 } } Invoke Actors with a delay set in milliseconds: iex&gt; SpawnSdk . invoke ( &quot;joe&quot; , system : &quot;spawn-system&quot; , command : &quot;ping&quot; , delay : 5_000 ) { :ok , :async } Invoke Actors scheduled to a specific DateTime: iex&gt; SpawnSdk . invoke ( &quot;joe&quot; , system : &quot;spawn-system&quot; , command : &quot;ping&quot; , scheduled_to : ~U[2023-01-01 00:32:00.145Z] ) { :ok , :async } Invoke Pooled Actors: iex&gt; SpawnSdk . invoke ( &quot;pooled_actor&quot; , system : &quot;spawn-system&quot; , command : &quot;ping&quot; , pooled : true ) { :ok , nil }","ref":"readme.html#client-api-examples","title":"Spawn Elixir SDK - Client API Examples","type":"extras"}]